import codecs
from typing import ClassVar
from unittest.mock import patch

from beets.dbcore.query import TrueQuery
from beets.importer import Action
from beets.library import Item
from beets.test.helper import (
    AutotagImportTestCase,
    AutotagStub,
    BeetsTestCase,
    IOMixin,
    PluginMixin,
    TerminalImportMixin,
)
from beetsplug.edit import EditPlugin, dump, load


class ModifyFileMocker:
    """Helper for modifying a file, replacing or editing its contents. Used for
    mocking the calls to the external editor during testing.
    """

    def __init__(self, contents=None, replacements=None):
        """`self.contents` and `self.replacements` are initialized here, in
        order to keep the rest of the functions of this class with the same
        signature as `EditPlugin.get_editor()`, making mocking easier.
            - `contents`: string with the contents of the file to be used for
            `overwrite_contents()`
            - `replacement`: dict with the in-place replacements to be used for
            `replace_contents()`, in the form {'previous string': 'new string'}

        TODO: check if it can be solved more elegantly with a decorator
        """
        self.contents = contents
        self.replacements = replacements
        self.action = self.overwrite_contents
        if replacements:
            self.action = self.replace_contents

    # The two methods below mock the `edit` utility function in the plugin.

    def overwrite_contents(self, filename, log):
        """Modify `filename`, replacing its contents with `self.contents`. If
        `self.contents` is empty, the file remains unchanged.
        """
        if self.contents:
            with codecs.open(filename, "w", encoding="utf-8") as f:
                f.write(self.contents)

    def replace_contents(self, filename, log):
        """Modify `filename`, reading its contents and replacing the strings
        specified in `self.replacements`.
        """
        with codecs.open(filename, "r", encoding="utf-8") as f:
            contents = f.read()
        for old, new_ in self.replacements.items():
            contents = contents.replace(old, new_)
        with codecs.open(filename, "w", encoding="utf-8") as f:
            f.write(contents)


class EditMixin(PluginMixin):
    """Helper containing some common functionality used for the Edit tests."""

    plugin = "edit"

    def assertItemFieldsModified(
        self, library_items, items, fields=[], allowed=["path"]
    ):
        """Assert that items in the library (`lib_items`) have different values
        on the specified `fields` (and *only* on those fields), compared to
        `items`.

        An empty `fields` list results in asserting that no modifications have
        been performed. `allowed` is a list of field changes that are ignored
        (they may or may not have changed; the assertion doesn't care).
        """
        for lib_item, item in zip(library_items, items):
            diff_fields = [
                field
                for field in lib_item._fields
                if lib_item[field] != item[field]
            ]
            assert set(diff_fields).difference(allowed) == set(fields)

    def run_mocked_interpreter(self, modify_file_args={}, stdin=[]):
        """Run the edit command during an import session, with mocked stdin and
        yaml writing.
        """
        m = ModifyFileMocker(**modify_file_args)
        with patch("beetsplug.edit.edit", side_effect=m.action):
            for char in stdin:
                self.importer.add_choice(char)
            self.importer.run()

    def run_mocked_command(self, modify_file_args={}, stdin=[], args=[]):
        """Run the edit command, with mocked stdin and yaml writing, and
        passing `args` to `run_command`."""
        m = ModifyFileMocker(**modify_file_args)
        with patch("beetsplug.edit.edit", side_effect=m.action):
            for char in stdin:
                self.io.addinput(char)
            self.run_command("edit", *args)


@patch("beets.library.Item.write")
class EditCommandTest(IOMixin, EditMixin, BeetsTestCase):
    """Black box tests for `beetsplug.edit`. Command line interaction is
    simulated using mocked stdin, and yaml editing via an external editor is
    simulated using `ModifyFileMocker`.
    """

    TRACK_COUNT = 10

    def setUp(self):
        super().setUp()
        # Add an album, storing the original fields for comparison.
        self.album = self.add_album_fixture(track_count=self.TRACK_COUNT)
        self.album_orig = {f: self.album[f] for f in self.album._fields}
        self.items_orig = [
            {f: item[f] for f in item._fields} for item in self.album.items()
        ]

    def test_title_edit_discard(self, mock_write):
        """Edit title for all items in the library, then discard changes."""
        # Edit track titles.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # Cancel.
            ["c"],
        )

        assert mock_write.call_count == 0
        self.assertItemFieldsModified(self.album.items(), self.items_orig, [])

    def test_title_edit_apply(self, mock_write):
        """Edit title for all items in the library, then apply changes."""
        # Edit track titles.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # Apply changes.
            ["a"],
        )

        assert mock_write.call_count == self.TRACK_COUNT
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["title", "mtime"]
        )

    def test_single_title_edit_apply(self, mock_write):
        """Edit title for one item in the library, then apply changes."""
        # Edit one track title.
        self.run_mocked_command(
            {"replacements": {"t\u00eftle 9": "modified t\u00eftle 9"}},
            # Apply changes.
            ["a"],
        )

        assert mock_write.call_count == 1
        # No changes except on last item.
        self.assertItemFieldsModified(
            list(self.album.items())[:-1], self.items_orig[:-1], []
        )
        assert list(self.album.items())[-1].title == "modified t\u00eftle 9"

    def test_title_edit_keep_editing_then_apply(self, mock_write):
        """Edit titles, keep editing once, then apply changes."""
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # keep Editing, then Apply
            ["e", "a"],
        )

        assert mock_write.call_count == self.TRACK_COUNT
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["title", "mtime"]
        )

    def test_title_edit_keep_editing_then_cancel(self, mock_write):
        """Edit titles, keep editing once, then cancel."""
        self.run_mocked_command(
            {"replacements": {"t\u00eftle": "modified t\u00eftle"}},
            # keep Editing, then Cancel
            ["e", "c"],
        )

        assert mock_write.call_count == 0
        self.assertItemFieldsModified(self.album.items(), self.items_orig, [])

    def test_noedit(self, mock_write):
        """Do not edit anything."""
        # Do not edit anything.
        self.run_mocked_command(
            {"contents": None},
            # No stdin.
            [],
        )

        assert mock_write.call_count == 0
        self.assertItemFieldsModified(self.album.items(), self.items_orig, [])

    def test_album_edit_apply(self, mock_write):
        """Edit the album field for all items in the library, apply changes.
        By design, the album should not be updated.""
        """
        # Edit album.
        self.run_mocked_command(
            {"replacements": {"\u00e4lbum": "modified \u00e4lbum"}},
            # Apply changes.
            ["a"],
        )

        assert mock_write.call_count == self.TRACK_COUNT
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["album", "mtime"]
        )
        # Ensure album is *not* modified.
        self.album.load()
        assert self.album.album == "\u00e4lbum"

    def test_single_edit_add_field(self, mock_write):
        """Edit the yaml file appending an extra field to the first item, then
        apply changes."""
        # Append "foo: bar" to item with id == 2. ("id: 1" would match both
        # "id: 1" and "id: 10")
        self.run_mocked_command(
            {"replacements": {"id: 2": "id: 2\nfoo: bar"}},
            # Apply changes.
            ["a"],
        )

        assert self.lib.items("id:2")[0].foo == "bar"
        # Even though a flexible attribute was written (which is not directly
        # written to the tags), write should still be called since templates
        # might use it.
        assert mock_write.call_count == 1

    def test_a_album_edit_apply(self, mock_write):
        """Album query (-a), edit album field, apply changes."""
        self.run_mocked_command(
            {"replacements": {"\u00e4lbum": "modified \u00e4lbum"}},
            # Apply changes.
            ["a"],
            args=["-a"],
        )

        self.album.load()
        assert mock_write.call_count == self.TRACK_COUNT
        assert self.album.album == "modified \u00e4lbum"
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["album", "mtime"]
        )

    def test_a_albumartist_edit_apply(self, mock_write):
        """Album query (-a), edit albumartist field, apply changes."""
        self.run_mocked_command(
            {"replacements": {"album artist": "modified album artist"}},
            # Apply changes.
            ["a"],
            args=["-a"],
        )

        self.album.load()
        assert mock_write.call_count == self.TRACK_COUNT
        assert self.album.albumartist == "the modified album artist"
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["albumartist", "mtime"]
        )

    def test_a_album_edit_preserves_missing_artpath(self, mock_write):
        """Album query (-a), edit album field, preserve missing artpath."""
        self.config["edit"]["albumfields"] = "album artpath"

        self.run_mocked_command(
            {"replacements": {"\u00e4lbum": "modified \u00e4lbum"}},
            # Apply changes.
            ["a"],
            args=["-a"],
        )

        self.album.load()
        assert mock_write.call_count == self.TRACK_COUNT
        assert self.album.album == "modified \u00e4lbum"
        assert self.album.artpath is None
        self.assertItemFieldsModified(
            self.album.items(), self.items_orig, ["album", "mtime"]
        )

    def test_malformed_yaml(self, mock_write):
        """Edit the yaml file incorrectly (resulting in a malformed yaml
        document)."""
        # Edit the yaml file to an invalid file.
        self.run_mocked_command(
            {"contents": "!MALFORMED"},
            # Edit again to fix? No.
            ["n"],
        )

        assert mock_write.call_count == 0

    def test_invalid_yaml(self, mock_write):
        """Edit the yaml file incorrectly (resulting in a well-formed but
        invalid yaml document)."""
        # Edit the yaml file to an invalid but parseable file.
        self.run_mocked_command(
            {"contents": "wellformed: yes, but invalid"},
            # No stdin.
            [],
        )

        assert mock_write.call_count == 0


class ApplyDataMatchingTest(PluginMixin, BeetsTestCase):
    """`apply_data` must match documents to objects by their ``id`` field,
    not by their position in the list. A reordered or otherwise misaligned
    document list must not cause one object's data to be applied to a
    different object.
    """

    plugin = "edit"

    def make_items(self):
        items = []
        for i in (1, 2, 3):
            item = Item(id=i, title=f"Title {i}", track=i, artist="Artist")
            items.append(item)
        return items

    def test_matches_by_id_when_documents_are_reordered(self):
        plugin = EditPlugin()
        items = self.make_items()
        old_data = [
            {"id": i.id, "title": i.title, "track": i.track} for i in items
        ]

        # The saved documents come back in a different order than `items`,
        # as could happen after a "keep editing" round-trip or a reordering
        # editor action. Only track 2's title was actually edited.
        new_data = [
            {"id": 3, "title": "Title 3", "track": 3},
            {"id": 1, "title": "Title 1", "track": 1},
            {"id": 2, "title": "Modified Title 2", "track": 2},
        ]

        plugin.apply_data(items, old_data, new_data)

        assert [i.title for i in items] == [
            "Title 1",
            "Modified Title 2",
            "Title 3",
        ]
        assert [i.track for i in items] == [1, 2, 3]

    def test_ignores_document_with_missing_id(self):
        plugin = EditPlugin()
        items = self.make_items()
        old_data = [
            {"id": i.id, "title": i.title, "track": i.track} for i in items
        ]

        # A document without an `id` must never be silently applied to an
        # unrelated object.
        new_data = [
            {"title": "Header-like doc", "track": 99},
            {"id": 2, "title": "Modified Title 2", "track": 2},
            {"id": 3, "title": "Title 3", "track": 3},
        ]

        plugin.apply_data(items, old_data, new_data)

        assert items[0].title == "Title 1"
        assert items[0].track == 1

    def test_ignores_documents_with_duplicate_id(self):
        plugin = EditPlugin()
        items = self.make_items()
        old_data = [
            {"id": i.id, "title": i.title, "track": i.track} for i in items
        ]

        # The user changed document 2's `id` to 1, an id that already
        # belongs to another document. Neither document should be applied:
        # we can't tell which one is legitimately item 1's data.
        new_data = [
            {"id": 1, "title": "Title 1", "track": 1},
            {"id": 1, "title": "Modified Title 2", "track": 2},
            {"id": 3, "title": "Title 3", "track": 3},
        ]

        plugin.apply_data(items, old_data, new_data)

        assert items[0].title == "Title 1"
        assert items[0].track == 1
        assert items[1].title == "Title 2"
        assert items[1].track == 2


class AlbumHeaderFieldsTest(PluginMixin, BeetsTestCase):
    """`_importer_edit_album_header` must strip item-only fixed fields
    (title, track, path, ...), since the header is built from a single
    item and then applied to every item in the album. Flexible fields
    are not part of either model's fixed schema, so they must be left
    alone rather than discarded by an overly broad filter.
    """

    plugin = "edit"

    class _StubTask:
        is_album = True

        def __init__(self, items):
            self.items = items

    def test_keeps_flexible_fields_but_drops_item_only_fields(self):
        item = Item(id=1, title="Title 1", track=1, album="Album", mood="Happy")
        task = self._StubTask([item])

        self.config["edit"]["albumfields"] = "album mood track title"

        header = EditPlugin()._importer_edit_album_header(task)

        assert header == {"album": "Album", "mood": "Happy"}


class EditDuringImporterTestCase(
    EditMixin, TerminalImportMixin, AutotagImportTestCase
):
    """TODO"""

    matching = AutotagStub.GOOD

    IGNORED: ClassVar[list[str]] = ["added", "album_id", "id", "mtime", "path"]

    def setUp(self):
        super().setUp()
        # Create some mediafiles, and store them for comparison.
        self.prepare_album_for_import(1)
        self.items_orig = [Item.from_path(f.path) for f in self.import_media]


class EditDuringImporterNonSingletonTest(EditDuringImporterTestCase):
    def setUp(self):
        super().setUp()
        self.importer = self.setup_importer()

    def test_importer_edit_album_header_and_items(self):
        """Edit both the album header and per-track fields simultaneously."""
        self.config["edit"]["itemfields"] = "title"
        self.config["edit"]["albumfields"] = "album"

        self.run_mocked_interpreter(
            {
                "replacements": {
                    "Tag Album": "Modified Album",
                    "Tag Track": "Modified Track",
                }
            },
            # eDit, Apply changes.
            ["d", "a"],
        )

        # All items should have the new album and new title.
        assert all(i.album == "Modified Album" for i in self.lib.items())
        assert all("Modified Track" in i.title for i in self.lib.items())
        assert self.lib.albums()[0].album == "Modified Album"

    def test_importer_edit_album_header_skip_no_albumfields(self):
        """When albumfields is empty, no header section is produced; editing
        works as before.
        """
        self.config["edit"]["itemfields"] = "title"
        self.config["edit"]["albumfields"] = ""

        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        assert all("Edited Track" in i.title for i in self.lib.items())

    def test_importer_edit_album_header_ignores_item_only_fields(self):
        """`albumfields` may be misconfigured with item-only fields (e.g.
        `track`, `title`, `path`) that vary per track. Those must not end
        up in the header, since the header gets applied to every item and
        would otherwise stamp one track's values onto the whole album.
        """
        self.prepare_album_for_import(3)
        self.items_orig = [Item.from_path(f.path) for f in self.import_media]

        self.config["edit"]["itemfields"] = "track title artist album"
        self.config["edit"]["albumfields"] = "album albumartist track title"

        self.run_mocked_interpreter(
            {"replacements": {"Tag Album": "Modified Album"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        titles = [i.title for i in self.lib.items()]
        tracks = [i.track for i in self.lib.items()]
        assert len(set(titles)) == len(titles)
        assert len(set(tracks)) == len(tracks)
        assert all(i.album == "Modified Album" for i in self.lib.items())

    def test_importer_edit_album_header_albumartist(self):
        """Edit albumartist in the header (default albumfields)."""
        self.run_mocked_interpreter(
            {"replacements": {"Tag Artist": "Modified Artist"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        assert all(
            i.albumartist is not None and "Modified Artist" in i.albumartist
            for i in self.lib.items()
        ) or all(
            i.albumartist is None and "Tag Artist" in i.artist
            for i in self.lib.items()
        )

    def test_importer_edit_album_header_reordered(self):
        """If the user moves the album header document below the track
        documents, it must still be recognized as the header (identified by
        the absence of an `id` field) rather than misapplying a track's
        fields to every item.
        """
        self.config["edit"]["itemfields"] = "title"
        self.config["edit"]["albumfields"] = "album"

        def reorder_and_edit(filename, log):
            with codecs.open(filename, encoding="utf-8") as f:
                docs = load(f.read())
            header = next(d for d in docs if "id" not in d)
            tracks = [d for d in docs if "id" in d]
            header["album"] = "Modified Album"
            with codecs.open(filename, "w", encoding="utf-8") as f:
                f.write(dump([*tracks, header]))

        with patch("beetsplug.edit.edit", side_effect=reorder_and_edit):
            self.importer.add_choice("d")
            self.importer.add_choice("a")
            self.importer.run()

        assert all(i.album == "Modified Album" for i in self.lib.items())
        assert all("Tag Track" in i.title for i in self.lib.items())

    def test_edit_apply_asis(self):
        """Edit the album field for all items in the library, apply changes,
        using the original item tags.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        # Check that only the 'title' field is modified.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            ["title", "albumartist", "albumartists"],
            [*self.IGNORED, "mb_albumartistid", "mb_albumartistids"],
        )
        assert all("Edited Track" in i.title for i in self.lib.items())

        # Ensure album is *not* fetched from a candidate.
        assert self.lib.albums()[0].mb_albumid == ""

    def test_edit_discard_asis(self):
        """Edit the album field for all items in the library, discard changes,
        using the original item tags.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Cancel, Use as-is.
            ["d", "c", "u"],
        )

        # Check that nothing is modified, the album is imported ASIS.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            [],
            [*self.IGNORED, "albumartist", "mb_albumartistid"],
        )
        assert all("Tag Track" in i.title for i in self.lib.items())

        # Ensure album is *not* fetched from a candidate.
        assert self.lib.albums()[0].mb_albumid == ""

    def test_edit_apply_candidate(self):
        """Edit the album field for all items in the library, apply changes,
        using a candidate.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes.
            ["c", "1", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_retag_apply(self):
        """Import the album using a candidate, then retag and edit and apply
        changes.
        """
        self.run_mocked_interpreter(
            {},
            # 1, Apply changes.
            ["1", Action.APPLY],
        )

        # Retag and edit track titles.  On retag, the importer will reset items
        # ids but not the db connections.
        self.importer.paths = []
        self.importer.query = TrueQuery()
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # eDit, Apply changes.
            ["d", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_discard_candidate(self):
        """Edit the album field for all items in the library, discard changes,
        using a candidate.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes.
            ["c", "1", "a"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())

        # Ensure album is fetched from a candidate.
        assert "albumid" in self.lib.albums()[0].mb_albumid

    def test_edit_apply_candidate_singleton(self):
        """Edit the album field for all items in the library, apply changes,
        using a candidate and singleton mode.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Applied Track": "Edited Track"}},
            # edit Candidates, 1, Apply changes, aBort.
            ["c", "1", "a", "b"],
        )

        # Check that 'title' field is modified, and other fields come from
        # the candidate.
        assert all("Edited Track " in i.title for i in self.lib.items())
        assert all("match " in i.mb_trackid for i in self.lib.items())


class EditDuringImporterSingletonTest(EditDuringImporterTestCase):
    def setUp(self):
        super().setUp()
        self.importer = self.setup_singleton_importer()

    def test_edit_apply_asis_singleton(self):
        """Edit the album field for all items in the library, apply changes,
        using the original item tags and singleton mode.
        """
        # Edit track titles.
        self.run_mocked_interpreter(
            {"replacements": {"Tag Track": "Edited Track"}},
            # eDit, Apply changes, aBort.
            ["d", "a", "b"],
        )

        # Check that only the 'title' field is modified.
        self.assertItemFieldsModified(
            self.lib.items(),
            self.items_orig,
            ["title"],
            [*self.IGNORED, "albumartist", "mb_albumartistid"],
        )
        assert all("Edited Track" in i.title for i in self.lib.items())



"""Adds Deezer release and track search support to the autotagger"""

from __future__ import annotations

import collections
import time
from typing import TYPE_CHECKING, ClassVar

import requests

from beets import config, ui
from beets.autotag import AlbumInfo, TrackInfo
from beets.dbcore import types
from beets.exceptions import UserError
from beets.metadata_plugins import IDResponse, SearchApiMetadataSourcePlugin

VARIOUS_ARTISTS_ID = 5080

if TYPE_CHECKING:
    from collections.abc import Sequence

    from beets.library import Item, Library
    from beets.metadata_plugins import QueryType, SearchParams

    from ._typing import JSONDict


class DeezerPlugin(SearchApiMetadataSourcePlugin[IDResponse]):
    item_types: ClassVar[dict[str, types.Type]] = {
        "deezer_track_rank": types.INTEGER,
        "deezer_track_id": types.INTEGER,
        "deezer_updated": types.DATE,
    }
    # Base URLs for the Deezer API
    # Documentation: https://developers.deezer.com/api/
    search_url = "https://api.deezer.com/search/"
    album_url = "https://api.deezer.com/album/"
    track_url = "https://api.deezer.com/track/"

    def __init__(self) -> None:
        super().__init__()

    def commands(self):
        """Add beet UI commands to interact with Deezer."""
        deezer_update_cmd = ui.Subcommand(
            "deezerupdate", help=f"Update {self.data_source} rank"
        )

        def func(lib: Library, opts, args):
            items = lib.items(args)
            self.deezerupdate(list(items), ui.should_write())

        deezer_update_cmd.func = func

        return [deezer_update_cmd]

    def album_for_id(self, album_id: str) -> AlbumInfo | None:
        """Fetch an album by its Deezer ID or URL."""
        if not (deezer_id := self._extract_id(album_id)):
            return None

        album_url = f"{self.album_url}{deezer_id}"
        if not (album_data := self.fetch_data(album_url)):
            return None

        contributors = album_data.get("contributors")
        if contributors is not None:
            artist, artist_id = self.get_artist(contributors)
        else:
            artist, artist_id = None, None

        release_date = album_data["release_date"]
        date_parts = [int(part) for part in release_date.split("-")]
        num_date_parts = len(date_parts)

        if num_date_parts == 3:
            year, month, day = date_parts
        elif num_date_parts == 2:
            year, month = date_parts
            day = None
        elif num_date_parts == 1:
            year = date_parts[0]
            month = None
            day = None
        else:
            raise UserError(
                f"Invalid `release_date` returned by {self.data_source} API: "
                f"{release_date!r}"
            )
        tracks_obj = self.fetch_data(f"{self.album_url}{deezer_id}/tracks")
        if tracks_obj is None:
            return None
        try:
            tracks_data = tracks_obj["data"]
        except KeyError:
            self._log.debug("Error fetching album tracks for {}", deezer_id)
            tracks_data = None
        if not tracks_data:
            return None
        while "next" in tracks_obj:
            tracks_obj = requests.get(tracks_obj["next"], timeout=10).json()
            tracks_data.extend(tracks_obj["data"])

        tracks = []
        medium_totals: dict[int | None, int] = collections.defaultdict(int)
        for i, track_data in enumerate(tracks_data, start=1):
            track = self._get_track(track_data)
            track.index = i
            medium_totals[track.medium] += 1
            tracks.append(track)
        for track in tracks:
            track.medium_total = medium_totals[track.medium]

        is_va = str(album_data["artist"]["id"]) == str(VARIOUS_ARTISTS_ID)
        if is_va:
            va_name = config["va_name"].as_str()
            artist = va_name

        return AlbumInfo(
            album=album_data["title"],
            album_id=deezer_id,
            deezer_album_id=deezer_id,
            artist=artist,
            artist_credit=(
                artist if is_va else self.get_artist([album_data["artist"]])[0]
            ),
            artist_id=str(artist_id),
            tracks=tracks,
            albumtype=album_data["record_type"],
            va=is_va,
            year=year,
            month=month,
            day=day,
            label=album_data["label"],
            mediums=max(filter(None, medium_totals.keys())),
            data_source=self.data_source,
            data_url=album_data["link"],
            cover_art_url=album_data.get("cover_xl"),
        )

    def track_for_id(self, track_id: str) -> TrackInfo | None:
        """Fetch a track by its Deezer ID or URL and return a
        TrackInfo object or None if the track is not found.

        :param track_id: (Optional) Deezer ID or URL for the track. Either
            ``track_id`` or ``track_data`` must be provided.

        """
        if not (deezer_id := self._extract_id(track_id)):
            self._log.debug("Invalid Deezer track_id: {}", track_id)
            return None

        if not (track_data := self.fetch_data(f"{self.track_url}{deezer_id}")):
            self._log.debug("Track not found: {}", track_id)
            return None

        track = self._get_track(track_data)

        # Get album's tracks to set `track.index` (position on the entire
        # release) and `track.medium_total` (total number of tracks on
        # the track's disc).
        if not (
            album_tracks_obj := self.fetch_data(
                f"{self.album_url}{track_data['album']['id']}/tracks"
            )
        ):
            return None

        try:
            album_tracks_data = album_tracks_obj["data"]
        except KeyError:
            self._log.debug(
                "Error fetching album tracks for {}", track_data["album"]["id"]
            )
            return None
        medium_total = 0
        for i, track_data in enumerate(album_tracks_data, start=1):
            if track_data["disk_number"] == track.medium:
                medium_total += 1
                if track_data["id"] == track.track_id:
                    track.index = i
        track.medium_total = medium_total
        return track

    def _get_track(self, track_data: JSONDict) -> TrackInfo:
        """Convert a Deezer track object dict to a TrackInfo object.

        :param track_data: Deezer Track object dict
        """
        artist, artist_id = self.get_artist(
            track_data.get("contributors", [track_data["artist"]])
        )
        return TrackInfo(
            title=track_data["title"],
            track_id=track_data["id"],
            deezer_track_id=track_data["id"],
            isrc=track_data.get("isrc"),
            artist=artist,
            artist_id=str(artist_id),
            length=track_data["duration"],
            index=track_data.get("track_position"),
            medium=track_data.get("disk_number"),
            deezer_track_rank=track_data.get("rank"),
            medium_index=track_data.get("track_position"),
            data_source=self.data_source,
            data_url=track_data["link"],
            deezer_updated=time.time(),
        )

    def get_search_query_with_filters(
        self,
        query_type: QueryType,
        items: Sequence[Item],
        artist: str,
        name: str,
        va_likely: bool,
    ) -> tuple[str, dict[str, str]]:
        query = f'album:"{name}"' if query_type == "album" else name
        if query_type == "track" or not va_likely:
            query += f' artist:"{artist}"'

        return query, {}

    def get_search_response(self, params: SearchParams) -> list[IDResponse]:
        """Search Deezer and return the raw result payload entries."""

        response = requests.get(
            f"{self.search_url}{params.query_type}",
            params={
                **params.filters,
                "q": params.query,
                "limit": str(params.limit),
            },
            timeout=10,
        )
        response.raise_for_status()
        return response.json()["data"]

    def deezerupdate(self, items: Sequence[Item], write: bool):
        """Obtain rank information from Deezer."""
        for index, item in enumerate(items, start=1):
            self._log.info(
                "Processing {}/{} tracks - {} ", index, len(items), item
            )
            try:
                deezer_track_id = item.deezer_track_id
            except AttributeError:
                self._log.debug("No deezer_track_id present for: {}", item)
                continue
            try:
                rank = self.fetch_data(
                    f"{self.track_url}{deezer_track_id}"
                ).get("rank")
                self._log.debug(
                    "Deezer track: {} has {} rank", deezer_track_id, rank
                )
            except Exception as e:
                self._log.debug("Invalid Deezer track_id: {}", e)
                continue
            item.deezer_track_rank = int(rank)
            item.store()
            item.deezer_updated = time.time()
            if write:
                item.try_write()

    def fetch_data(self, url: str):
        try:
            response = requests.get(url, timeout=10)
            response.raise_for_status()
            data = response.json()
        except requests.exceptions.RequestException as e:
            self._log.error("Error fetching data from {}\n Error: {}", url, e)
            return None
        if "error" in data:
            self._log.debug("Deezer API error: {}", data["error"]["message"])
            return None
        return data



"""Dataclasses for managing artist credits and tracklists from Discogs."""

from __future__ import annotations

import re
from dataclasses import asdict, dataclass, field
from functools import cached_property
from typing import TYPE_CHECKING, NamedTuple

from beets import config
from beets.util import unique_list

from .types import ArtistInfo

if TYPE_CHECKING:
    from confuse import ConfigView

    from beets.autotag import TrackInfo

    from . import DiscogsPlugin
    from .types import Artist, Track, TracklistInfo

DISAMBIGUATION_RE = re.compile(r" \(\d+\)")


@dataclass
class ArtistState:
    """Represent Discogs artist credits.

    This object centralizes the plugin's policy for which Discogs artist fields
    to prefer (name vs. ANV), how to treat 'Various', how to format join
    phrases, and how to separate featured artists. It exposes both per-artist
    components and fully joined strings for common tag targets like 'artist' and
    'artist_credit'.
    """

    class ValidArtist(NamedTuple):
        """A normalized, render-ready artist entry extracted from Discogs data.

        Instances represent the subset of Discogs artist information needed for
        tagging, including the join token following the artist and the
        lowercased Discogs role string (e.g. "featuring", "written-by", "remix")
        """

        id: str
        name: str
        credit: str
        join: str
        role: str

        def __hash__(self) -> int:
            return hash(self.id)

        def __eq__(self, other: object) -> bool:
            if isinstance(other, self.__class__):
                return self.id == other.id

            return False

        def get_artist(self, property_name: str) -> str:
            """Return the requested display field with its trailing join token.

            The join token is normalized so commas become ', ' and other join
            phrases are surrounded with spaces, producing a single fragment that
            can be concatenated to form a full artist string.
            """
            join = {",": ", ", "": ""}.get(self.join, f" {self.join} ")
            return f"{getattr(self, property_name)}{join}"

    raw_artists: list[Artist]
    use_anv: bool
    use_credit_anv: bool
    featured_string: str
    should_strip_disambiguation: bool

    @property
    def info(self) -> ArtistInfo:
        """Expose the state in the shape expected by downstream tag mapping."""
        return {k: getattr(self, k) for k in ArtistInfo.__annotations__}  # type: ignore[return-value]

    def strip_disambiguation(self, text: str) -> str:
        """Strip Discogs disambiguation suffixes from an artist or label string.

        This removes Discogs-specific numeric suffixes like 'Name (5)' and can
        be applied to multi-artist strings as well (e.g., 'A (1) & B (2)'). When
        the feature is disabled, the input is returned unchanged.
        """
        if self.should_strip_disambiguation:
            return DISAMBIGUATION_RE.sub("", text)
        return text

    @cached_property
    def valid_artists(self) -> list[ValidArtist]:
        """Build the ordered, filtered list of artists used for rendering.

        The resulting list normalizes Discogs entries by:
        - substituting the configured 'Various Artists' name when Discogs uses
          'Various'
        - choosing between name and ANV according to plugin settings
        - capturing join tokens so the original credit formatting is preserved
        """
        va_name = config["va_name"].as_str()
        return [
            self.ValidArtist(
                str(a["id"]),
                self.strip_disambiguation(anv if self.use_anv else name),
                self.strip_disambiguation(anv if self.use_credit_anv else name),
                a["join"].strip(),
                a["role"].lower(),
            )
            for a in self.raw_artists
            if (
                (name := va_name if a["name"] == "Various" else a["name"])
                and (anv := a["anv"] or name)
            )
        ]

    @property
    def main_artists(self) -> list[ValidArtist]:
        """Return the per-artist display names used for the 'artist' field."""
        artists = [
            a for a in self.valid_artists if not a.role or "featuring" in a.role
        ]
        return unique_list(artists)

    @property
    def artists_ids(self) -> list[str]:
        """Return Discogs artist IDs for the main artists, preserving order."""
        return [a.id for a in self.main_artists]

    @property
    def artist_id(self) -> str:
        """Return the primary Discogs artist ID."""
        return self.artists_ids[0]

    @property
    def artists(self) -> list[str]:
        """Return the per-artist display names used for the 'artist' field."""
        return [a.name for a in self.main_artists]

    @property
    def artists_credit(self) -> list[str]:
        """Return the per-artist display names used for the credit field."""
        return [a.credit for a in self.main_artists]

    @property
    def artist(self) -> str:
        """Return the fully rendered artist string using display names."""
        return self.join_artists("name")

    @property
    def artist_credit(self) -> str:
        """Return the fully rendered artist credit string."""
        return self.join_artists("credit")

    @property
    def lyricists(self) -> list[str] | None:
        return [
            a.name for a in self.valid_artists if "lyrics" in a.role
        ] or None

    @property
    def arrangers(self) -> list[str] | None:
        return [
            a.name for a in self.valid_artists if "arrang" in a.role
        ] or None

    @property
    def remixers(self) -> list[str] | None:
        return [a.name for a in self.valid_artists if "remix" in a.role] or None

    @property
    def composers(self) -> list[str] | None:
        return [
            a.name for a in self.valid_artists if "written-by" in a.role
        ] or None

    def join_artists(self, property_name: str) -> str:
        """Render a single artist string with join phrases and featured artists.

        Non-featured artists are concatenated using their join tokens. Featured
        artists are appended after the configured 'featured' marker, preserving
        Discogs order while keeping featured credits separate from the main
        artist string.
        """
        non_featured = [a for a in self.main_artists if not a.role]
        featured = [a for a in self.main_artists if "featuring" in a.role]

        artist = "".join(a.get_artist(property_name) for a in non_featured)
        if featured:
            if "feat" not in artist:
                artist += f" {self.featured_string} "

            artist += ", ".join(a.get_artist(property_name) for a in featured)

        return artist

    @classmethod
    def from_config(
        cls,
        config: ConfigView,
        artists: list[Artist],
        for_album_artist: bool = False,
    ) -> ArtistState:
        return cls(
            artists,
            config["anv"]["album_artist" if for_album_artist else "artist"].get(
                bool
            ),
            config["anv"]["artist_credit"].get(bool),
            config["featured_string"].as_str(),
            config["strip_disambiguation"].get(bool),
        )


@dataclass
class TracklistState:
    index: int = 0
    index_tracks: dict[int, str] = field(default_factory=dict)
    tracks: list[TrackInfo] = field(default_factory=list)
    divisions: list[str] = field(default_factory=list)
    next_divisions: list[str] = field(default_factory=list)
    mediums: list[str | None] = field(default_factory=list)
    medium_indices: list[str | None] = field(default_factory=list)

    @property
    def info(self) -> TracklistInfo:
        return asdict(self)  # type: ignore[return-value]

    @classmethod
    def build(
        cls,
        plugin: DiscogsPlugin,
        clean_tracklist: list[Track],
        albumartistinfo: ArtistState,
    ) -> TracklistState:
        state = cls()
        for track in clean_tracklist:
            if track["type_"] == "track":
                state.index += 1
                if state.next_divisions:
                    state.divisions += state.next_divisions
                    state.next_divisions.clear()
                track_info, medium, medium_index = plugin.get_track_info(
                    track, state.index, state.divisions, albumartistinfo
                )
                track_info.track_alt = track["position"]
                state.tracks.append(track_info)
                state.mediums.append(medium or None)
                state.medium_indices.append(medium_index or None)
            else:
                state.next_divisions.append(track["title"])
                try:
                    state.divisions.pop()
                except IndexError:
                    pass
                state.index_tracks[state.index + 1] = track["title"]
        return state



"""Use command-line tools to check for audio file corruption."""

import errno
import os
import shlex
import sys
from subprocess import STDOUT, CalledProcessError, check_output, list2cmdline
from typing import Literal

import confuse

from beets import config, importer, ui
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand
from beets.util import displayable_path, par_map
from beets.util.color import colorize


class CheckerCommandError(Exception):
    """Raised when running a checker failed.

    Attributes:
        checker: Checker command name.
        path: Path to the file being validated.
        errno: Error number from the checker execution error.
        msg: Message from the checker execution error.
    """

    def __init__(self, cmd, oserror):
        self.checker = cmd[0]
        self.path = cmd[-1]
        self.errno = oserror.errno
        self.msg = str(oserror)


class BadFiles(BeetsPlugin):
    def __init__(self):
        super().__init__()
        self.verbose = False

        self.config.add(
            {"import_action_on_warning": "ask", "import_action_on_error": "ask"}
        )

        self.register_listener("import_task_start", self.on_import_task_start)
        self.register_listener(
            "import_task_before_choice", self.on_import_task_before_choice
        )

    def run_command(self, cmd):
        self._log.debug(
            "running command: {}", displayable_path(list2cmdline(cmd))
        )
        try:
            output = check_output(cmd, stderr=STDOUT)
            errors = 0
            status = 0
        except CalledProcessError as e:
            output = e.output
            errors = 1
            status = e.returncode
        except OSError as e:
            raise CheckerCommandError(cmd, e)
        output = output.decode(sys.getdefaultencoding(), "replace")
        return status, errors, [line for line in output.split("\n") if line]

    def check_mp3val(self, path):
        status, errors, output = self.run_command(["mp3val", path])
        if status == 0:
            output = [line for line in output if line.startswith("WARNING:")]
            errors = len(output)
        return status, errors, output

    def check_flac(self, path):
        return self.run_command(["flac", "-wst", path])

    def check_custom(self, command):
        def checker(path):
            cmd = shlex.split(command)
            cmd.append(path)
            return self.run_command(cmd)

        return checker

    def get_checker(self, ext):
        ext = ext.lower()
        try:
            command = self.config["commands"].get(dict).get(ext)
        except confuse.NotFoundError:
            command = None
        if command:
            return self.check_custom(command)
        if ext == "mp3":
            return self.check_mp3val
        if ext == "flac":
            return self.check_flac
        return None

    def check_item(self, item):
        # First, check whether the path exists. If not, the user
        # should probably run `beet update` to cleanup your library.
        dpath = displayable_path(item.path)
        self._log.debug("checking path: {}", dpath)
        if not os.path.exists(item.path):
            ui.print_(f"{colorize('text_error', dpath)}: file does not exist")

        # Run the checker against the file if one is found
        ext = os.path.splitext(item.path)[1][1:].decode("utf8", "ignore")
        checker = self.get_checker(ext)
        if not checker:
            self._log.error("no checker specified in the config for {}", ext)
            return []
        path = item.path
        if not isinstance(path, str):
            path = item.path.decode(sys.getfilesystemencoding())
        try:
            status, errors, output = checker(path)
        except CheckerCommandError as e:
            if e.errno == errno.ENOENT:
                self._log.error(
                    "command not found: {0.checker} when validating file: {0.path}",
                    e,
                )
            else:
                self._log.error("error invoking {0.checker}: {0.msg}", e)
            return []

        error_lines = []

        if status > 0:
            error_lines.append(
                f"{colorize('text_error', dpath)}: checker exited with status {status}"
            )
            for line in output:
                error_lines.append(f"  {line}")

        elif errors > 0:
            error_lines.append(
                f"{colorize('text_warning', dpath)}: checker found"
                f" {errors} errors or warnings"
            )
            for line in output:
                error_lines.append(f"  {line}")
        elif self.verbose:
            error_lines.append(f"{colorize('text_success', dpath)}: ok")

        return error_lines

    def on_import_task_start(self, task, session):
        if not self.config["check_on_import"].get(False):
            return

        checks_failed = []

        for item in task.items:
            error_lines = self.check_item(item)
            if error_lines:
                checks_failed.append(error_lines)

        if checks_failed:
            task._badfiles_checks_failed = checks_failed

    def handle_import_action(
        self,
        action: Literal["abort", "skip", "continue"],
        failure_type: Literal["error", "warning"],
    ) -> importer.Action | None:
        action_name_by_action = {
            "abort": "Aborting",
            "skip": "Skipping",
            "continue": "Continuing",
        }
        ui.print_(
            f"{ui.colorize('text_warning', action_name_by_action[action])}"
            f" due to import_action_on_{failure_type} configuration"
        )
        if action == "abort":
            raise importer.ImportAbortError()
        if action == "skip":
            return importer.Action.SKIP
        return None

    def on_import_task_before_choice(self, task, session):
        if hasattr(task, "_badfiles_checks_failed"):
            actions = confuse.Choice(["ask", "abort", "skip", "continue"])
            warning_action = self.config["import_action_on_warning"].get(
                actions
            )
            error_action = self.config["import_action_on_error"].get(actions)

            ui.print_(
                f"{colorize('text_warning', 'BAD')} one or more files failed checks:"
            )

            found_warning = False
            found_error = False
            for error in task._badfiles_checks_failed:
                for error_line in error:
                    if (
                        "checker found 0 errors or warnings"
                        in error_line.lower()
                    ):
                        continue

                    if "warning" in error_line.lower():
                        found_warning = True
                    if "error" in error_line.lower():
                        found_error = True

                    ui.print_(error_line)

            # Check for and handle automatic actions.
            # Errors always take precedence over warnings.
            if found_error and error_action != "ask":
                return self.handle_import_action(error_action, "error")
            if found_warning and warning_action != "ask":
                return self.handle_import_action(warning_action, "warning")

            # Defer the quiet check to after automatic import action options are handled
            if config["import"]["quiet"]:
                return None

            ui.print_()
            ui.print_("What would you like to do?")

            sel = ui.input_options(["aBort", "skip", "continue"])

            if sel == "s":
                return importer.Action.SKIP
            if sel == "c":
                return None
            if sel == "b":
                raise importer.ImportAbortError()
            raise Exception(f"Unexpected selection: {sel}")
        return None

    def command(self, lib, opts, args):
        # Get items from arguments
        items = lib.items(args)
        self.verbose = opts.verbose

        def check_and_print(item):
            for error_line in self.check_item(item):
                ui.print_(error_line)

        par_map(check_and_print, items)

    def commands(self):
        bad_command = Subcommand(
            "bad", help="check for corrupt or missing files"
        )
        bad_command.parser.add_option(
            "-v",
            "--verbose",
            action="store_true",
            default=False,
            dest="verbose",
            help="view results for both the bad and uncorrupted files",
        )
        bad_command.func = self.command
        return [bad_command]



# TODO: Tests in this fire are very bad. Stop using Mocks in this module.
from __future__ import annotations

import os
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, Mock, PropertyMock

import pytest

from beets import config
from beets.dbcore.sort import FixedFieldSort, MultipleSort, NullSort
from beets.library import Album, Item, parse_query_string
from beets.test._common import item
from beets.test.helper import BeetsTestCase, IOMixin, PathsMixin, PluginTestCase
from beets.ui import UserError
from beets.util import CHAR_REPLACE, syspath
from beetsplug.smartplaylist import SmartPlaylistPlugin

if TYPE_CHECKING:
    from pathlib import Path

_p = pytest.param


class PlaylistDirMixin(PathsMixin):
    @property
    def playlist_dir(self) -> Path:
        return self.temp_path / "playlists"


class SmartPlaylistTest(PlaylistDirMixin, BeetsTestCase):
    def test_build_queries(self):
        spl = SmartPlaylistPlugin()
        assert spl._matched_playlists == set()
        assert spl._unmatched_playlists == set()

        config["smartplaylist"]["playlists"].set([])
        spl.build_queries()
        assert spl._matched_playlists == set()
        assert spl._unmatched_playlists == set()

        config["smartplaylist"]["playlists"].set(
            [
                {"name": "foo", "query": "FOO foo"},
                {"name": "bar", "album_query": ["BAR bar1", "BAR bar2"]},
                {"name": "baz", "query": "BAZ baz", "album_query": "BAZ baz"},
            ]
        )
        spl.build_queries()
        assert spl._matched_playlists == set()
        foo_foo = parse_query_string("FOO foo", Item)
        baz_baz = parse_query_string("BAZ baz", Item)
        baz_baz2 = parse_query_string("BAZ baz", Album)
        # Multiple queries are now stored as a tuple of (query, sort) tuples
        bar_queries = tuple(
            [
                parse_query_string("BAR bar1", Album),
                parse_query_string("BAR bar2", Album),
            ]
        )
        assert spl._unmatched_playlists == {
            ("foo", foo_foo, (None, None)),
            ("baz", baz_baz, baz_baz2),
            ("bar", (None, None), (bar_queries, None)),
        }

    def test_build_queries_with_sorts(self):
        spl = SmartPlaylistPlugin()
        config["smartplaylist"]["playlists"].set(
            [
                {"name": "no_sort", "query": "foo"},
                {"name": "one_sort", "query": "foo year+"},
                {"name": "only_empty_sorts", "query": ["foo", "bar"]},
                {"name": "one_non_empty_sort", "query": ["foo year+", "bar"]},
                {
                    "name": "multiple_sorts",
                    "query": ["foo year+", "bar genres-"],
                },
                {
                    "name": "mixed",
                    "query": ["foo year+", "bar", "baz genres+ id-"],
                },
            ]
        )

        spl.build_queries()

        # Multiple queries now return a tuple of (query, sort) tuples, not combined
        sorts = {}
        for name, (query_data, sort), _ in spl._unmatched_playlists:
            if isinstance(query_data, tuple):
                # Tuple of queries - each has its own sort
                sorts[name] = [s for _, s in query_data]
            else:
                sorts[name] = sort

        sort = FixedFieldSort  # short cut since we're only dealing with this
        assert sorts["no_sort"] == NullSort()
        assert sorts["one_sort"] == sort("year")
        # Multiple queries store individual sorts in the tuple
        assert all(isinstance(x, NullSort) for x in sorts["only_empty_sorts"])
        assert sorts["one_non_empty_sort"] == [sort("year"), NullSort()]
        assert sorts["multiple_sorts"] == [sort("year"), sort("genres", False)]
        assert sorts["mixed"] == [
            sort("year"),
            NullSort(),
            MultipleSort([sort("genres"), sort("id", False)]),
        ]

    def test_matches(self):
        spl = SmartPlaylistPlugin()

        a = MagicMock(Album)
        i = MagicMock(Item)

        assert not spl.matches(i, None, None)
        assert not spl.matches(a, None, None)

        query = Mock()
        query.match.side_effect = {i: True}.__getitem__
        assert spl.matches(i, query, None)
        assert not spl.matches(a, query, None)

        a_query = Mock()
        a_query.match.side_effect = {a: True}.__getitem__
        assert not spl.matches(i, None, a_query)
        assert spl.matches(a, None, a_query)

        assert spl.matches(i, query, a_query)
        assert spl.matches(a, query, a_query)

        # Test with list of queries
        q1 = Mock()
        q1.match.return_value = False
        q2 = Mock()
        q2.match.side_effect = {i: True}.__getitem__
        queries_list = [(q1, None), (q2, None)]
        assert spl.matches(i, queries_list, None)
        assert not spl.matches(a, queries_list, None)

    def test_db_changes(self):
        spl = SmartPlaylistPlugin()

        nones = None, None
        pl1 = "1", ("q1", None), nones
        pl2 = "2", ("q2", None), nones
        pl3 = "3", ("q3", None), nones

        spl._unmatched_playlists = {pl1, pl2, pl3}
        spl._matched_playlists = set()

        spl.matches = Mock(return_value=False)
        spl.db_change(None, "nothing")
        assert spl._unmatched_playlists == {pl1, pl2, pl3}
        assert spl._matched_playlists == set()

        spl.matches.side_effect = lambda _, q, __: q == "q3"
        spl.db_change(None, "matches 3")
        assert spl._unmatched_playlists == {pl1, pl2}
        assert spl._matched_playlists == {pl3}

        spl.matches.side_effect = lambda _, q, __: q == "q1"
        spl.db_change(None, "matches 3")
        assert spl._matched_playlists == {pl1, pl3}
        assert spl._unmatched_playlists == {pl2}

    def test_playlist_update(self):
        spl = SmartPlaylistPlugin()

        i = Mock(path=b"/tagada.mp3")
        i.evaluate_template.side_effect = lambda pl, *_: os.fsdecode(
            pl
        ).replace("$title", "ta:ga:da")

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = {pl}

        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = str(self.playlist_dir)
        spl.update_playlists(lib)

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = self.playlist_dir / "ta_ga_da-my_playlist_.m3u"
        assert m3u_filepath.exists()
        content = m3u_filepath.read_bytes()

        assert content == b"/tagada.mp3\n"

    def test_playlist_update_output_extm3u(self):
        spl = SmartPlaylistPlugin()

        i = MagicMock()
        type(i).artist = PropertyMock(return_value="fake artist")
        type(i).title = PropertyMock(return_value="fake title")
        type(i).length = PropertyMock(return_value=300.123)
        type(i).path = PropertyMock(return_value=b"/tagada.mp3")
        i.evaluate_template.side_effect = lambda pl, *_: os.fsdecode(
            pl
        ).replace("$title", "ta:ga:da")

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = {pl}

        config["smartplaylist"]["output"] = "extm3u"
        config["smartplaylist"]["prefix"] = "http://beets:8337/files"
        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = str(self.playlist_dir)
        spl.update_playlists(lib)

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = self.playlist_dir / "ta_ga_da-my_playlist_.m3u"
        assert m3u_filepath.exists()
        content = m3u_filepath.read_bytes()

        assert content == (
            b"#EXTM3U\n"
            b"#EXTINF:300,fake artist - fake title\n"
            b"http://beets:8337/files/tagada.mp3\n"
        )

    def test_playlist_update_output_extm3u_fields(self):
        spl = SmartPlaylistPlugin()

        i = MagicMock()
        type(i).artist = PropertyMock(return_value="Fake Artist")
        type(i).title = PropertyMock(return_value="fake Title")
        type(i).length = PropertyMock(return_value=300.123)
        type(i).path = PropertyMock(return_value=b"/tagada.mp3")
        a = {"id": 456, "genres": ["Rock", "Pop"]}
        i.__getitem__.side_effect = a.__getitem__
        i.evaluate_template.side_effect = lambda pl, *_: os.fsdecode(
            pl
        ).replace("$title", "ta:ga:da")

        lib = Mock()
        lib.replacements = CHAR_REPLACE
        lib.items.return_value = [i]
        lib.albums.return_value = []

        q = Mock()
        a_q = Mock()
        pl = b"$title-my<playlist>.m3u", (q, None), (a_q, None)
        spl._matched_playlists = {pl}

        config["smartplaylist"]["output"] = "extm3u"
        config["smartplaylist"]["relative_to"] = False
        config["smartplaylist"]["playlist_dir"] = str(self.playlist_dir)
        config["smartplaylist"]["fields"] = ["id", "genres"]
        spl.update_playlists(lib)

        lib.items.assert_called_once_with(q, None)
        lib.albums.assert_called_once_with(a_q, None)

        m3u_filepath = self.playlist_dir / "ta_ga_da-my_playlist_.m3u"
        assert m3u_filepath.exists()
        content = m3u_filepath.read_bytes()

        assert content == (
            b"#EXTM3U\n"
            b'#EXTINF:300 id="456" genres="Rock%3B%20Pop",Fake Artist - fake Title\n'
            b"/tagada.mp3\n"
        )

    def test_get_playlist_items(self):
        """Test get playlist items.

        - Items preserve their order in the playlist
        - There are no duplicates when items match multiple queries
        """
        self.add_item(path=b"/item1.mp3", id=1)
        self.add_item(path=b"/item2.mp3", id=2)
        self.add_item(path=b"/item3.mp3", id=3)
        queries_and_sorts = (("path::item id-", None), ("path::item3", None))

        actual_items = SmartPlaylistPlugin.get_playlist_items(
            self.lib, (queries_and_sorts, None), (None, None)
        )

        assert [i.id for i in actual_items] == [3, 2, 1]


class TestGetItemURI:
    @pytest.fixture
    def plugin(self, config, plugin_config):
        plugin_config = {"prefix": "http://beets:8337/files", **plugin_config}
        config["smartplaylist"].set(plugin_config)

        return SmartPlaylistPlugin()

    @pytest.fixture(autouse=True)
    def _setup(self, monkeypatch):
        monkeypatch.setattr(Item, "destination", lambda _: b"/tagada.mp3")

    @pytest.fixture
    def item(self):
        return item(
            id=3,
            artist="fake artist",
            title="fake title",
            length=300.123,
            path=b"/imported/path/with/dont/move/tagada.mp3",
        )

    @pytest.mark.parametrize(
        "plugin_config, expected_uri",
        [
            _p(
                {},
                b"http://beets:8337/files/imported/path/with/dont/move/tagada.mp3",
                id="path by default",
            ),
            _p(
                {"dest_regen": True},
                b"http://beets:8337/files/tagada.mp3",
                id="dest_regen uses item destination",
            ),
            _p(
                {
                    "uri_format": "http://beets:8337/item/$id/file",
                    "dest_regen": True,
                },
                b"http://beets:8337/item/3/file",
                id="uri_format takes precedence",
            ),
        ],
    )
    def test_get_item_uri(self, plugin, item, expected_uri):
        assert plugin.get_item_uri(item) == expected_uri


class SmartPlaylistCLITest(PlaylistDirMixin, IOMixin, PluginTestCase):
    plugin = "smartplaylist"

    def setUp(self):
        super().setUp()

        self.item = self.add_item()
        config["smartplaylist"]["playlists"].set(
            [
                {"name": "my_playlist.m3u", "query": self.item.title},
                {"name": "all.m3u", "query": ""},
            ]
        )
        config["smartplaylist"]["playlist_dir"] = str(self.playlist_dir)

    def test_splupdate(self):
        with pytest.raises(UserError):
            self.run_with_output("splupdate", "tagada")

        self.run_with_output("splupdate", "my_playlist")
        m3u_path = self.playlist_dir / "my_playlist.m3u"
        assert m3u_path.exists()
        assert m3u_path.read_bytes() == self.item.path + b"\n"
        os.remove(syspath(m3u_path))

        self.run_with_output("splupdate", "my_playlist.m3u")
        assert m3u_path.read_bytes() == self.item.path + b"\n"
        os.remove(syspath(m3u_path))

        self.run_with_output("splupdate")
        for name in ("my_playlist.m3u", "all.m3u"):
            assert (
                self.playlist_dir / name
            ).read_bytes() == self.item.path + b"\n"

    def test_splupdate_unknown_playlist_error_is_sorted_and_quoted(self):
        config["smartplaylist"]["playlists"].set(
            [
                {"name": "z last.m3u", "query": self.item.title},
                {"name": "rock'n roll.m3u", "query": self.item.title},
                {"name": "a one.m3u", "query": self.item.title},
            ]
        )

        with pytest.raises(UserError) as exc_info:
            self.run_with_output("splupdate", "tagada")

        assert str(exc_info.value) == (
            "No playlist matching any of "
            "'a one.m3u' 'rock'\"'\"'n roll.m3u' 'z last.m3u' found"
        )

    def test_splupdate_log_output(self):
        with self.assertLogs("beets.smartplaylist", level="INFO") as logs:
            self.run_with_output("splupdate", "my_playlist")

        output = "\n".join(logs.output)
        assert "Updating 1 smart playlists..." in output
        assert "Creating playlist my_playlist.m3u: 1 tracks." in output
        assert "1 playlists updated" in output

    def test_splupdate_verbose_log_output(self):
        with self.assertLogs("beets.smartplaylist", level="DEBUG") as logs:
            self.run_with_output("splupdate", "my_playlist")

        output = "\n".join(logs.output)
        assert "Updating 1 smart playlists..." in output
        assert "Creating playlist my_playlist.m3u: 1 tracks." in output
        assert "the ärtist - " in output
        assert "1 playlists updated" in output

    def test_splupdate_pretend_log_output(self):
        with self.assertLogs("beets.smartplaylist", level="INFO") as logs:
            self.run_with_output("splupdate", "--pretend", "my_playlist")

        output = "\n".join(logs.output)
        assert "Updating 1 smart playlists..." in output
        assert "Creating playlist my_playlist.m3u: 1 tracks." in output
        assert "1 playlists would be updated" in output



"""An AURA server using Flask."""

from __future__ import annotations

import os
import re
from dataclasses import dataclass
from mimetypes import guess_type
from typing import TYPE_CHECKING, ClassVar

from flask import (
    Blueprint,
    Flask,
    current_app,
    make_response,
    request,
    send_file,
)
from typing_extensions import Self

from beets import config
from beets.dbcore.query import AndQuery, MatchQuery, NotQuery, RegexpQuery
from beets.dbcore.sort import FixedFieldSort, MultipleSort, SlowFieldSort
from beets.library import Album, Item
from beets.plugins import BeetsPlugin
from beets.ui import Subcommand, _open_library

if TYPE_CHECKING:
    from collections.abc import Mapping

    from beets.dbcore.query import SQLiteType
    from beets.library import LibModel, Library

# Constants

# AURA server information
# TODO: Add version information
SERVER_INFO = {
    "aura-version": "0",
    "server": "beets-aura",
    "server-version": "0.1",
    "auth-required": False,
    "features": ["albums", "artists", "images"],
}

# Maps AURA Track attribute to beets Item attribute
TRACK_ATTR_MAP = {
    # Required
    "title": "title",
    "artist": "artist",
    # Optional
    "album": "album",
    "track": "track",  # Track number on album
    "tracktotal": "tracktotal",
    "disc": "disc",
    "disctotal": "disctotal",
    "year": "year",
    "month": "month",
    "day": "day",
    "bpm": "bpm",
    "genre": "genres",
    "genres": "genres",
    "recording-mbid": "mb_trackid",  # beets trackid is MB recording
    "track-mbid": "mb_releasetrackid",
    "composer": "composers",
    "albumartist": "albumartist",
    "comments": "comments",
    # Optional for Audio Metadata
    # TODO: Support the mimetype attribute, format != mime type
    # "mimetype": track.format,
    "duration": "length",
    "framerate": "samplerate",
    # I don't think beets has a framecount field
    # "framecount": ???,
    "channels": "channels",
    "bitrate": "bitrate",
    "bitdepth": "bitdepth",
    "size": "filesize",
}

# Maps AURA Album attribute to beets Album attribute
ALBUM_ATTR_MAP = {
    # Required
    "title": "album",
    "artist": "albumartist",
    # Optional
    "tracktotal": "albumtotal",
    "disctotal": "disctotal",
    "year": "year",
    "month": "month",
    "day": "day",
    "genre": "genres",
    "genres": "genres",
    "release-mbid": "mb_albumid",
    "release-group-mbid": "mb_releasegroupid",
}

# Maps AURA Artist attribute to beets Item field
# Artists are not first-class in beets, so information is extracted from
# beets Items.
ARTIST_ATTR_MAP = {
    # Required
    "name": "artist",
    # Optional
    "artist-mbid": "mb_artistid",
}


@dataclass
class AURADocument:
    """Base class for building AURA documents."""

    model_cls: ClassVar[type[LibModel]]

    lib: Library
    args: Mapping[str, str]

    @classmethod
    def from_app(cls) -> Self:
        """Initialise the document using the global app and request."""
        return cls(current_app.config["lib"], request.args)

    @staticmethod
    def error(status, title, detail):
        """Make a response for an error following the JSON:API spec.

        Args:
            status: An HTTP status code string, e.g. "404 Not Found".
            title: A short, human-readable summary of the problem.
            detail: A human-readable explanation specific to this
                occurrence of the problem.
        """
        document = {
            "errors": [{"status": status, "title": title, "detail": detail}]
        }
        return make_response(document, status)

    @classmethod
    def get_attribute_converter(cls, beets_attr: str) -> type[SQLiteType]:
        """Work out what data type an attribute should be for beets.

        Args:
            beets_attr: The name of the beets attribute, e.g. "title".
        """
        try:
            # Look for field in list of Album fields
            # and get python type of database type.
            # See beets.library.Album and beets.dbcore.types
            return cls.model_cls._fields[beets_attr].model_type
        except KeyError:
            # Fall back to string (NOTE: probably not good)
            return str

    def translate_filters(self):
        """Translate filters from request arguments to a beets Query."""
        # The format of each filter key in the request parameter is:
        # filter[<attribute>]. This regex extracts <attribute>.
        pattern = re.compile(r"filter\[(?P<attribute>[a-zA-Z0-9_-]+)\]")
        queries = []
        for key, value in self.args.items():
            match = pattern.match(key)
            if match:
                # Extract attribute name from key
                aura_attr = match.group("attribute")
                # Get the beets version of the attribute name
                beets_attr = self.attribute_map.get(aura_attr, aura_attr)
                converter = self.get_attribute_converter(beets_attr)
                value = converter(value)
                # Add exact match query to list
                # Use a slow query so it works with all fields
                queries.append(
                    self.model_cls.field_query(beets_attr, value, MatchQuery)
                )
        # NOTE: AURA doesn't officially support multiple queries
        return AndQuery(queries)

    def translate_sorts(self, sort_arg):
        """Translate an AURA sort parameter into a beets Sort.

        Args:
            sort_arg: The value of the 'sort' query parameter; a comma
                separated list of fields to sort by, in order.
                E.g. "-year,title".
        """
        # Change HTTP query parameter to a list
        aura_sorts = sort_arg.strip(",").split(",")
        sorts = []
        for aura_attr in aura_sorts:
            if aura_attr[0] == "-":
                ascending = False
                # Remove leading "-"
                aura_attr = aura_attr[1:]
            else:
                # JSON:API default
                ascending = True
            # Get the beets version of the attribute name
            beets_attr = self.attribute_map.get(aura_attr, aura_attr)
            # Use slow sort so it works with all fields (inc. computed)
            sorts.append(SlowFieldSort(beets_attr, ascending=ascending))
        return MultipleSort(sorts)

    def paginate(self, collection):
        """Get a page of the collection and the URL to the next page.

        Args:
            collection: The raw data from which resource objects can be
                built. Could be an sqlite3.Cursor object (tracks and
                albums) or a list of strings (artists).
        """
        # Pages start from zero
        page = self.args.get("page", 0, int)
        # Use page limit defined in config by default.
        default_limit = config["aura"]["page_limit"].get(int)
        limit = self.args.get("limit", default_limit, int)
        # start = offset of first item to return
        start = page * limit
        # end = offset of last item + 1
        end = start + limit
        if end > len(collection):
            end = len(collection)
            next_url = None
        else:
            # Not the last page so work out links.next url
            if not self.args:
                # No existing arguments, so current page is 0
                next_url = f"{request.url}?page=1"
            elif not self.args.get("page", None):
                # No existing page argument, so add one to the end
                next_url = f"{request.url}&page=1"
            else:
                # Increment page token by 1
                next_url = request.url.replace(
                    f"page={page}", f"page={page + 1}"
                )
        # Get only the items in the page range
        data = [
            self.get_resource_object(self.lib, collection[i])
            for i in range(start, end)
        ]
        return data, next_url

    def get_included(self, data, include_str):
        """Build a list of resource objects for inclusion.

        Args:
            data: An array of dicts in the form of resource objects.
            include_str: A comma separated list of resource types to
                include. E.g. "tracks,images".
        """
        # Change HTTP query parameter to a list
        to_include = include_str.strip(",").split(",")
        # Build a list of unique type and id combinations
        # For each resource object in the primary data, iterate over it's
        # relationships. If a relationship matches one of the types
        # requested for inclusion (e.g. "albums") then add each type-id pair
        # under the "data" key to unique_identifiers, checking first that
        # it has not already been added. This ensures that no resources are
        # included more than once.
        unique_identifiers = []
        for res_obj in data:
            for rel_name, rel_obj in res_obj["relationships"].items():
                if rel_name in to_include:
                    # NOTE: Assumes relationship is to-many
                    for identifier in rel_obj["data"]:
                        if identifier not in unique_identifiers:
                            unique_identifiers.append(identifier)
        # TODO: I think this could be improved
        included = []
        for identifier in unique_identifiers:
            res_type = identifier["type"]
            if res_type == "track":
                track_id = int(identifier["id"])
                track = self.lib.get_item(track_id)
                included.append(
                    TrackDocument.get_resource_object(self.lib, track)
                )
            elif res_type == "album":
                album_id = int(identifier["id"])
                album = self.lib.get_album(album_id)
                included.append(
                    AlbumDocument.get_resource_object(self.lib, album)
                )
            elif res_type == "artist":
                artist_id = identifier["id"]
                included.append(
                    ArtistDocument.get_resource_object(self.lib, artist_id)
                )
            elif res_type == "image":
                image_id = identifier["id"]
                included.append(
                    ImageDocument.get_resource_object(self.lib, image_id)
                )
            else:
                raise ValueError(f"Invalid resource type: {res_type}")
        return included

    def all_resources(self):
        """Build document for /tracks, /albums or /artists."""
        query = self.translate_filters()
        sort_arg = self.args.get("sort", None)
        if sort_arg:
            sort = self.translate_sorts(sort_arg)
            # For each sort field add a query which ensures all results
            # have a non-empty, non-zero value for that field.
            query.subqueries.extend(
                NotQuery(
                    self.model_cls.field_query(s.field, "(^$|^0$)", RegexpQuery)
                )
                for s in sort.sorts
            )
        else:
            sort = None
        # Get information from the library
        collection = self.get_collection(query=query, sort=sort)
        # Convert info to AURA form and paginate it
        data, next_url = self.paginate(collection)
        document = {"data": data}
        # If there are more pages then provide a way to access them
        if next_url:
            document["links"] = {"next": next_url}
        # Include related resources for each element in "data"
        include_str = self.args.get("include", None)
        if include_str:
            document["included"] = self.get_included(data, include_str)
        return document

    def single_resource_document(self, resource_object):
        """Build document for a specific requested resource.

        Args:
            resource_object: A dictionary in the form of a JSON:API
                resource object.
        """
        document = {"data": resource_object}
        include_str = self.args.get("include", None)
        if include_str:
            # [document["data"]] is because arg needs to be list
            document["included"] = self.get_included(
                [document["data"]], include_str
            )
        return document


class TrackDocument(AURADocument):
    """Class for building documents for /tracks endpoints."""

    model_cls = Item

    attribute_map = TRACK_ATTR_MAP

    def get_collection(self, query=None, sort=None):
        """Get Item objects from the library.

        Args:
            query: A beets Query object or a beets query string.
            sort: A beets Sort object.
        """
        return self.lib.items(query, sort)

    @classmethod
    def get_attribute_converter(cls, beets_attr: str) -> type[SQLiteType]:
        """Work out what data type an attribute should be for beets.

        Args:
            beets_attr: The name of the beets attribute, e.g. "title".
        """
        # filesize is a special field (read from disk not db?)
        if beets_attr == "filesize":
            return int

        return super().get_attribute_converter(beets_attr)

    @staticmethod
    def get_resource_object(lib: Library, track):
        """Construct a JSON:API resource object from a beets Item.

        Args:
            track: A beets Item object.
        """
        attributes = {}
        # Use aura => beets attribute map, e.g. size => filesize
        for aura_attr, beets_attr in TRACK_ATTR_MAP.items():
            a = getattr(track, beets_attr)
            # Only set attribute if it's not None, 0, "", etc.
            # NOTE: This could result in required attributes not being set
            if a:
                attributes[aura_attr] = a

        # JSON:API one-to-many relationship to parent album
        relationships = {
            "artists": {"data": [{"type": "artist", "id": track.artist}]}
        }
        # Only add album relationship if not singleton
        if not track.singleton:
            relationships["albums"] = {
                "data": [{"type": "album", "id": str(track.album_id)}]
            }

        return {
            "type": "track",
            "id": str(track.id),
            "attributes": attributes,
            "relationships": relationships,
        }

    def single_resource(self, track_id):
        """Get track from the library and build a document.

        Args:
            track_id: The beets id of the track (integer).
        """
        track = self.lib.get_item(track_id)
        if not track:
            return self.error(
                "404 Not Found",
                "No track with the requested id.",
                f"There is no track with an id of {track_id} in the library.",
            )
        return self.single_resource_document(
            self.get_resource_object(self.lib, track)
        )


class AlbumDocument(AURADocument):
    """Class for building documents for /albums endpoints."""

    model_cls = Album

    attribute_map = ALBUM_ATTR_MAP

    def get_collection(self, query=None, sort=None):
        """Get Album objects from the library.

        Args:
            query: A beets Query object or a beets query string.
            sort: A beets Sort object.
        """
        return self.lib.albums(query, sort)

    @staticmethod
    def get_resource_object(lib: Library, album):
        """Construct a JSON:API resource object from a beets Album.

        Args:
            album: A beets Album object.
        """
        attributes = {}
        # Use aura => beets attribute name map
        for aura_attr, beets_attr in ALBUM_ATTR_MAP.items():
            a = getattr(album, beets_attr)
            # Only set attribute if it's not None, 0, "", etc.
            # NOTE: This could mean required attributes are not set
            if a:
                attributes[aura_attr] = a

        # Get beets Item objects for all tracks in the album sorted by
        # track number. Sorting is not required but it's nice.
        query = MatchQuery("album_id", album.id)
        sort = FixedFieldSort("track", ascending=True)
        tracks = lib.items(query, sort)
        # JSON:API one-to-many relationship to tracks on the album
        relationships = {
            "tracks": {
                "data": [{"type": "track", "id": str(t.id)} for t in tracks]
            }
        }
        # Add images relationship if album has associated images
        if album.artpath:
            path = os.fsdecode(album.artpath)
            filename = path.split("/")[-1]
            image_id = f"album-{album.id}-{filename}"
            relationships["images"] = {
                "data": [{"type": "image", "id": image_id}]
            }
        # Add artist relationship if artist name is same on tracks
        # Tracks are used to define artists so don't albumartist
        # Check for all tracks in case some have featured artists
        if album.albumartist in [t.artist for t in tracks]:
            relationships["artists"] = {
                "data": [{"type": "artist", "id": album.albumartist}]
            }

        return {
            "type": "album",
            "id": str(album.id),
            "attributes": attributes,
            "relationships": relationships,
        }

    def single_resource(self, album_id):
        """Get album from the library and build a document.

        Args:
            album_id: The beets id of the album (integer).
        """
        album = self.lib.get_album(album_id)
        if not album:
            return self.error(
                "404 Not Found",
                "No album with the requested id.",
                f"There is no album with an id of {album_id} in the library.",
            )
        return self.single_resource_document(
            self.get_resource_object(self.lib, album)
        )


class ArtistDocument(AURADocument):
    """Class for building documents for /artists endpoints."""

    model_cls = Item

    attribute_map = ARTIST_ATTR_MAP

    def get_collection(self, query=None, sort=None):
        """Get a list of artist names from the library.

        Args:
            query: A beets Query object or a beets query string.
            sort: A beets Sort object.
        """
        # Gets only tracks with matching artist information
        tracks = self.lib.items(query, sort)
        collection = []
        for track in tracks:
            # Do not add duplicates
            if track.artist not in collection:
                collection.append(track.artist)
        return collection

    @staticmethod
    def get_resource_object(lib: Library, artist_id):
        """Construct a JSON:API resource object for the given artist.

        Args:
            artist_id: A string which is the artist's name.
        """
        # Get tracks where artist field exactly matches artist_id
        query = MatchQuery("artist", artist_id)
        tracks = lib.items(query)
        if not tracks:
            return None

        # Get artist information from the first track
        # NOTE: It could be that the first track doesn't have a
        # MusicBrainz id but later tracks do, which isn't ideal.
        attributes = {}
        # Use aura => beets attribute map, e.g. artist => name
        for aura_attr, beets_attr in ARTIST_ATTR_MAP.items():
            a = getattr(tracks[0], beets_attr)
            # Only set attribute if it's not None, 0, "", etc.
            # NOTE: This could mean required attributes are not set
            if a:
                attributes[aura_attr] = a

        relationships = {
            "tracks": {
                "data": [{"type": "track", "id": str(t.id)} for t in tracks]
            }
        }
        album_query = MatchQuery("albumartist", artist_id)
        albums = lib.albums(query=album_query)
        if len(albums) != 0:
            relationships["albums"] = {
                "data": [{"type": "album", "id": str(a.id)} for a in albums]
            }

        return {
            "type": "artist",
            "id": artist_id,
            "attributes": attributes,
            "relationships": relationships,
        }

    def single_resource(self, artist_id):
        """Get info for the requested artist and build a document.

        Args:
            artist_id: A string which is the artist's name.
        """
        artist_resource = self.get_resource_object(self.lib, artist_id)
        if not artist_resource:
            return self.error(
                "404 Not Found",
                "No artist with the requested id.",
                f"There is no artist with an id of {artist_id} in the library.",
            )
        return self.single_resource_document(artist_resource)


def safe_filename(fn):
    """Check whether a string is a simple (non-path) filename.

    For example, `foo.txt` is safe because it is a "plain" filename. But
    `foo/bar.txt` and `../foo.txt` and `.` are all non-safe because they
    can traverse to other directories other than the current one.
    """
    # Rule out any directories.
    if os.path.basename(fn) != fn:
        return False

    # In single names, rule out Unix directory traversal names.
    if fn in (".", ".."):
        return False

    return True


class ImageDocument(AURADocument):
    """Class for building documents for /images/(id) endpoints."""

    model_cls = Album

    @staticmethod
    def get_image_path(lib: Library, image_id):
        """Works out the full path to the image with the given id.

        Returns None if there is no such image.

        Args:
            image_id: A string in the form
                "<parent_type>-<parent_id>-<img_filename>".
        """
        # Split image_id into its constituent parts
        id_split = image_id.split("-")
        if len(id_split) < 3:
            # image_id is not in the required format
            return None
        parent_type = id_split[0]
        parent_id = id_split[1]
        img_filename = "-".join(id_split[2:])
        if not safe_filename(img_filename):
            return None

        # Get the path to the directory parent's images are in
        if parent_type == "album":
            album = lib.get_album(int(parent_id))
            if not album or not album.artpath:
                return None
            # Cut the filename off of artpath
            # This is in preparation for supporting images in the same
            # directory that are not tracked by beets.
            artpath = os.fsdecode(album.artpath)
            dir_path = "/".join(artpath.split("/")[:-1])
        else:
            # Images for other resource types are not supported
            return None

        img_path = os.path.join(dir_path, img_filename)
        # Check the image actually exists
        if os.path.isfile(img_path):
            return img_path
        return None

    @staticmethod
    def get_resource_object(lib: Library, image_id):
        """Construct a JSON:API resource object for the given image.

        Args:
            image_id: A string in the form
                "<parent_type>-<parent_id>-<img_filename>".
        """
        # Could be called as a static method, so can't use
        # self.get_image_path()
        image_path = ImageDocument.get_image_path(lib, image_id)
        if not image_path:
            return None

        attributes = {
            "role": "cover",
            "mimetype": guess_type(image_path)[0],
            "size": os.path.getsize(image_path),
        }
        try:
            from PIL import Image
        except ImportError:
            pass
        else:
            im = Image.open(image_path)
            attributes["width"] = im.width
            attributes["height"] = im.height

        relationships = {}
        # Split id into [parent_type, parent_id, filename]
        id_split = image_id.split("-")
        relationships[f"{id_split[0]}s"] = {
            "data": [{"type": id_split[0], "id": id_split[1]}]
        }

        return {
            "id": image_id,
            "type": "image",
            # Remove attributes that are None, 0, "", etc.
            "attributes": {k: v for k, v in attributes.items() if v},
            "relationships": relationships,
        }

    def single_resource(self, image_id):
        """Get info for the requested image and build a document.

        Args:
            image_id: A string in the form
                "<parent_type>-<parent_id>-<img_filename>".
        """
        image_resource = self.get_resource_object(self.lib, image_id)
        if not image_resource:
            return self.error(
                "404 Not Found",
                "No image with the requested id.",
                f"There is no image with an id of {image_id} in the library.",
            )
        return self.single_resource_document(image_resource)


# Initialise flask blueprint
aura_bp = Blueprint("aura_bp", __name__)


@aura_bp.route("/server")
def server_info():
    """Respond with info about the server."""
    return {"data": {"type": "server", "id": "0", "attributes": SERVER_INFO}}


# Track endpoints


@aura_bp.route("/tracks")
def all_tracks():
    """Respond with a list of all tracks and related information."""
    return TrackDocument.from_app().all_resources()


@aura_bp.route("/tracks/<int:track_id>")
def single_track(track_id):
    """Respond with info about the specified track.

    Args:
        track_id: The id of the track provided in the URL (integer).
    """
    return TrackDocument.from_app().single_resource(track_id)


@aura_bp.route("/tracks/<int:track_id>/audio")
def audio_file(track_id):
    """Supply an audio file for the specified track.

    Args:
        track_id: The id of the track provided in the URL (integer).
    """
    track = current_app.config["lib"].get_item(track_id)
    if not track:
        return AURADocument.error(
            "404 Not Found",
            "No track with the requested id.",
            f"There is no track with an id of {track_id} in the library.",
        )

    path = os.fsdecode(track.path)
    if not os.path.isfile(path):
        return AURADocument.error(
            "404 Not Found",
            "No audio file for the requested track.",
            f"There is no audio file for track {track_id} at the expected"
            " location",
        )

    file_mimetype = guess_type(path)[0]
    if not file_mimetype:
        return AURADocument.error(
            "500 Internal Server Error",
            "Requested audio file has an unknown mimetype.",
            f"The audio file for track {track_id} has an unknown mimetype. "
            f"Its file extension is {path.split('.')[-1]}.",
        )

    # Check that the Accept header contains the file's mimetype
    # Takes into account */* and audio/*
    # Adding support for the bitrate parameter would require some effort so I
    # left it out. This means the client could be sent an error even if the
    # audio doesn't need transcoding.
    if not request.accept_mimetypes.best_match([file_mimetype]):
        return AURADocument.error(
            "406 Not Acceptable",
            "Unsupported MIME type or bitrate parameter in Accept header.",
            f"The audio file for track {track_id} is only available as"
            f" {file_mimetype} and bitrate parameters are not supported.",
        )

    return send_file(
        path,
        mimetype=file_mimetype,
        # Handles filename in Content-Disposition header
        as_attachment=True,
        # Tries to upgrade the stream to support range requests
        conditional=True,
    )


# Album endpoints


@aura_bp.route("/albums")
def all_albums():
    """Respond with a list of all albums and related information."""
    return AlbumDocument.from_app().all_resources()


@aura_bp.route("/albums/<int:album_id>")
def single_album(album_id):
    """Respond with info about the specified album.

    Args:
        album_id: The id of the album provided in the URL (integer).
    """
    return AlbumDocument.from_app().single_resource(album_id)


# Artist endpoints
# Artist ids are their names


@aura_bp.route("/artists")
def all_artists():
    """Respond with a list of all artists and related information."""
    return ArtistDocument.from_app().all_resources()


# Using the path converter allows slashes in artist_id
@aura_bp.route("/artists/<path:artist_id>")
def single_artist(artist_id):
    """Respond with info about the specified artist.

    Args:
        artist_id: The id of the artist provided in the URL. A string
            which is the artist's name.
    """
    return ArtistDocument.from_app().single_resource(artist_id)


# Image endpoints
# Image ids are in the form <parent_type>-<parent_id>-<img_filename>
# For example: album-13-cover.jpg


@aura_bp.route("/images/<string:image_id>")
def single_image(image_id):
    """Respond with info about the specified image.

    Args:
        image_id: The id of the image provided in the URL. A string in
            the form "<parent_type>-<parent_id>-<img_filename>".
    """
    return ImageDocument.from_app().single_resource(image_id)


@aura_bp.route("/images/<string:image_id>/file")
def image_file(image_id):
    """Supply an image file for the specified image.

    Args:
        image_id: The id of the image provided in the URL. A string in
            the form "<parent_type>-<parent_id>-<img_filename>".
    """
    img_path = ImageDocument.get_image_path(current_app.config["lib"], image_id)
    if not img_path:
        return AURADocument.error(
            "404 Not Found",
            "No image with the requested id.",
            f"There is no image with an id of {image_id} in the library",
        )
    return send_file(img_path)


# WSGI app


def create_app():
    """An application factory for use by a WSGI server."""
    config["aura"].add(
        {
            "host": "127.0.0.1",
            "port": 8337,
            "cors": [],
            "cors_supports_credentials": False,
            "page_limit": 500,
        }
    )

    app = Flask(__name__)
    # Register AURA blueprint view functions under a URL prefix
    app.register_blueprint(aura_bp, url_prefix="/aura")
    # AURA specifies mimetype MUST be this
    app.config["JSONIFY_MIMETYPE"] = "application/vnd.api+json"
    # Disable auto-sorting of JSON keys
    app.config["JSON_SORT_KEYS"] = False
    # Provide a way to access the beets library
    # The normal method of using the Library and config provided in the
    # command function is not used because create_app() could be called
    # by an external WSGI server.
    # NOTE: this uses a 'private' function from beets.ui.__init__
    app.config["lib"] = _open_library(config)

    # Enable CORS if required
    cors = config["aura"]["cors"].as_str_seq(list)
    if cors:
        from flask_cors import CORS

        # "Accept" is the only header clients use
        app.config["CORS_ALLOW_HEADERS"] = "Accept"
        app.config["CORS_RESOURCES"] = {r"/aura/*": {"origins": cors}}
        app.config["CORS_SUPPORTS_CREDENTIALS"] = config["aura"][
            "cors_supports_credentials"
        ].get(bool)
        CORS(app)

    return app


# Beets Plugin Hook


class AURAPlugin(BeetsPlugin):
    """The BeetsPlugin subclass for the AURA server plugin."""

    def __init__(self):
        """Add configuration options for the AURA plugin."""
        super().__init__()

    def commands(self):
        """Add subcommand used to run the AURA server."""

        def run_aura(lib, opts, args):
            """Run the application using Flask's built in-server.

            Args:
                lib: A beets Library object (not used).
                opts: Command line options. An optparse.Values object.
                args: The list of arguments to process (not used).
            """
            app = create_app()
            # Start the built-in server (not intended for production)
            app.run(
                host=self.config["host"].get(str),
                port=self.config["port"].get(int),
                debug=opts.debug,
                threaded=True,
            )

        run_aura_cmd = Subcommand("aura", help="run an AURA server")
        run_aura_cmd.parser.add_option(
            "-d",
            "--debug",
            action="store_true",
            default=False,
            help="use Flask debug mode",
        )
        run_aura_cmd.func = run_aura
        return [run_aura_cmd]



"""Tests for discogs plugin."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import pytest
from discogs_client import Client, Release

from beets import config
from beets.library import Item
from beets.test.helper import TestHelper
from beetsplug.discogs import ArtistState, DiscogsPlugin

from .factories import discogs as factories

if TYPE_CHECKING:
    from beetsplug.discogs import types


_p = pytest.param


def _artist(name: str, **kwargs):
    return {
        "id": 1,
        "name": name,
        "join": "",
        "role": "",
        "anv": "",
        "tracks": "",
        "resource_url": "",
    } | kwargs


def audio_track(
    title: str, position: str = "", duration: str = "", **kwargs
) -> types.AudioTrack:
    kwargs.update(title=title, position=position, duration=duration)
    return factories.AudioTrackFactory.build(**kwargs)


def index_track(
    title: str, sub_tracks: list[types.AudioTrack], **kwargs
) -> types.IndexTrack:
    kwargs.update(title=title, sub_tracks=sub_tracks)
    return factories.IndexTrackFactory.build(**kwargs)


def heading_track(title: str, **kwargs) -> types.HeadingTrack:
    kwargs.update(title=title)
    return factories.HeadingTrackFactory.build(**kwargs)


def get_release(data: dict[str, Any]) -> Release:
    return Release(Client("doesn't matter"), data)


@pytest.fixture(autouse=True)
def _patch_discogs_setup(monkeypatch):
    """Autouse fixture to patch DiscogsPlugin.setup for each test."""
    monkeypatch.setattr(
        "beetsplug.discogs.DiscogsPlugin.setup", lambda *_: None
    )


class DiscogsTestMixin:
    @pytest.fixture
    def plugin_config(self):
        return {}

    @pytest.fixture
    def plugin(self, plugin_config):
        plugin = DiscogsPlugin()
        plugin.config.set(plugin_config)
        return plugin

    def _make_release(self, tracks=None):
        """Return discogs_client.Release.

        The returned object is incomplete, including just the fields required
        for tests in this module.
        """
        data = {
            "id": 11111111,
            "uri": "https://www.discogs.com/release/111111111",
            "title": "ALBUM TITLE",
            "year": 3001,
            "artists": [_artist("ARTIST NAME", join=",")],
            "formats": [
                {
                    "descriptions": ["FORMAT DESC 1", "FORMAT DESC 2"],
                    "name": "FORMAT",
                    "qty": 1,
                }
            ],
            # genres and styles are reversed in Discogs
            "genres": ["STYLE1", "STYLE2"],
            "styles": ["GENRE1", "GENRE2"],
            "labels": [{"name": "LABEL NAME", "catno": "CATALOG NUMBER"}],
            "tracklist": tracks or [],
        }

        return get_release(data)

    def _make_release_from_positions(self, positions):
        """Return discogs_client.Release with tracks at the given positions."""
        tracks = [
            audio_track(f"TITLE{i}", position)
            for (i, position) in enumerate(positions, start=1)
        ]
        return self._make_release(tracks)


class TestDGAlbumInfo(DiscogsTestMixin, TestHelper):
    def test_parse_media_for_tracks(self):
        tracks = [
            audio_track("TITLE ONE", "1", "01:01"),
            audio_track("TITLE TWO", "2", "02:02"),
        ]
        release = self._make_release(tracks=tracks)

        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks
        assert d.media == "FORMAT"
        assert t[0].media == d.media
        assert t[1].media == d.media

    def test_parse_medium_numbers_single_medium(self):
        release = self._make_release_from_positions(["1", "2"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 1
        assert t[0].medium == 1
        assert t[0].medium_total == 2
        assert t[1].medium == 1
        assert t[0].medium_total == 2

    def test_parse_medium_numbers_two_mediums(self):
        release = self._make_release_from_positions(["1-1", "2-1"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 2
        assert t[0].medium == 1
        assert t[0].medium_total == 1
        assert t[1].medium == 2
        assert t[1].medium_total == 1

    def test_parse_medium_numbers_two_mediums_two_sided(self):
        release = self._make_release_from_positions(["A1", "B1", "C1"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 2
        assert t[0].medium == 1
        assert t[0].medium_total == 2
        assert t[0].medium_index == 1
        assert t[1].medium == 1
        assert t[1].medium_total == 2
        assert t[1].medium_index == 2
        assert t[2].medium == 2
        assert t[2].medium_total == 1
        assert t[2].medium_index == 1

    def test_parse_track_indices(self):
        release = self._make_release_from_positions(["1", "2"])
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[0].medium_total == 2
        assert t[1].medium_index == 2
        assert t[1].index == 2
        assert t[1].medium_total == 2

    def test_parse_track_indices_several_media(self):
        release = self._make_release_from_positions(
            ["1-1", "1-2", "2-1", "3-1"]
        )
        d = DiscogsPlugin().get_album_info(release)
        t = d.tracks

        assert d.mediums == 3
        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[0].medium_total == 2
        assert t[1].medium_index == 2
        assert t[1].index == 2
        assert t[1].medium_total == 2
        assert t[2].medium_index == 1
        assert t[2].index == 3
        assert t[2].medium_total == 1
        assert t[3].medium_index == 1
        assert t[3].index == 4
        assert t[3].medium_total == 1

    def test_parse_minimal_release(self):
        """Test parsing of a release with the minimal amount of information."""
        data = {
            "id": 123,
            "uri": "https://www.discogs.com/release/123456-something",
            "tracklist": [audio_track("A", "1", "01:01")],
            "artists": [_artist("ARTIST NAME", id=321)],
            "title": "TITLE",
        }
        release = get_release(data)
        d = DiscogsPlugin().get_album_info(release)
        assert d.artist == "ARTIST NAME"
        assert d.album == "TITLE"
        assert len(d.tracks) == 1

    def test_default_genre_style_settings(self):
        """Test genre default settings, genres to genre, styles to style"""
        release = self._make_release_from_positions(["1", "2"])

        d = DiscogsPlugin().get_album_info(release)
        assert d.style == "STYLE1, STYLE2"
        assert d.genres == ["GENRE1", "GENRE2"]

    def test_append_style_to_genre(self):
        """Test appending style to genre if config enabled"""
        config["discogs"]["append_style_genre"] = True
        release = self._make_release_from_positions(["1", "2"])

        d = DiscogsPlugin().get_album_info(release)
        assert d.style == "STYLE1, STYLE2"
        assert d.genres == ["GENRE1", "GENRE2", "STYLE1", "STYLE2"]

    def test_append_style_to_genre_no_styles(self):
        """Test nothing appended to genre if style is empty"""
        config["discogs"]["append_style_genre"] = True
        release = self._make_release_from_positions(["1", "2"])
        release.data["genres"] = []

        d = DiscogsPlugin().get_album_info(release)
        assert d.style is None
        assert d.genres == ["GENRE1", "GENRE2"]


class TestStripDisambiguation(DiscogsTestMixin):
    @pytest.fixture
    def album_info(self, plugin):
        data = {
            "id": 123,
            "uri": "https://www.discogs.com/release/123456-something",
            "tracklist": [
                {
                    "title": "track",
                    "position": "A",
                    "type_": "track",
                    "duration": "5:44",
                    "artists": [_artist("TEST ARTIST (5)", id=11146)],
                }
            ],
            "artists": [
                _artist("ARTIST NAME (2)", id=321, join="&"),
                _artist("OTHER ARTIST (5)", id=322),
            ],
            "title": "title",
            "labels": [{"name": "LABEL NAME (5)", "catno": "catalog number"}],
        }
        return plugin.get_album_info(get_release(data))

    @pytest.mark.parametrize("plugin_config", [{"strip_disambiguation": True}])
    def test_strip_disambiguation(self, album_info):
        """Test removing disambiguation from all disambiguated fields."""
        assert album_info.artist == "ARTIST NAME & OTHER ARTIST"
        assert album_info.artists == ["ARTIST NAME", "OTHER ARTIST"]
        assert album_info.artists_ids == ["321", "322"]
        assert album_info.tracks[0].artist == "TEST ARTIST"
        assert album_info.tracks[0].artists == ["TEST ARTIST"]
        assert album_info.tracks[0].artist_id == "11146"
        assert album_info.tracks[0].artists_ids == ["11146"]
        assert album_info.label == "LABEL NAME"

    @pytest.mark.parametrize("plugin_config", [{"strip_disambiguation": False}])
    def test_dont_strip_disambiguation(self, album_info):
        """Test disabling disambiguation removal from all disambiguated fields."""
        assert album_info.artist == "ARTIST NAME (2) & OTHER ARTIST (5)"
        assert album_info.artists == ["ARTIST NAME (2)", "OTHER ARTIST (5)"]
        assert album_info.tracks[0].artist == "TEST ARTIST (5)"
        assert album_info.tracks[0].artists == ["TEST ARTIST (5)"]
        assert album_info.label == "LABEL NAME (5)"


class TestTracklist(DiscogsTestMixin):
    @pytest.mark.parametrize(
        "positions,expected_mediums,expected_tracks",
        [
            _p(["1", "2", "3"], 1, 3, id="without-sides"),
            _p(["A1", "A2", "B1", "B2"], 1, 4, id="with-sides"),
            _p(["A1", "A2", "B1", "C1"], 2, 4, id="multiple-lp"),
            _p(["1-1", "1-2", "2-1", "3-1"], 3, 4, id="multiple-cd"),
            _p(["I", "II", "III", "IV"], 1, 4, id="non-standard"),
            _p(["1", "2.1", "2.2", "3"], 1, 3, id="subtracks-dot"),
            _p(["A1", "A2.1", "A2.2", "A3"], 1, 3, id="subtracks-dot-with-sides"),
            _p(["A1", "A2a", "A2b", "A3"], 1, 3, id="subtracks-letter"),
            _p(["A1", "A2.a", "A2.b", "A3"], 1, 3, id="subtracks-letter-with-dot"),
            _p(["1.1", "1.2", "2.1", "2.2"], 1, 2, id="multiple-subtrack-groups"),
            _p(["1", "2", "Video 1"], 2, 3, id="subtracks-extra-material"),
        ],
    )  # fmt: skip
    def test_parse_tracklist_positions(
        self, plugin, positions, expected_mediums, expected_tracks
    ):
        release = self._make_release_from_positions(positions)
        d = plugin.get_album_info(release)

        assert d.mediums == expected_mediums
        assert len(d.tracks) == expected_tracks

    @pytest.mark.parametrize(
        "plugin_config,tracks,expected_mediums,expected_tracks",
        [
            _p(
                {"index_tracks": False},
                [
                    heading_track("MEDIUM TITLE"),
                    audio_track("TITLE ONE", "1.1"),
                    audio_track("TITLE TWO", "1.2"),
                ],
                1,
                [("TITLE ONE / TITLE TWO", "MEDIUM TITLE")],
                id="flat-logical-subtracks",
            ),
            _p(
                {"index_tracks": False},
                [
                    audio_track("TITLE ONE", "1"),
                    index_track(
                        "TRACK GROUP TITLE",
                        [
                            audio_track("SUBTITLE ONE", "2.1", "01:01"),
                            audio_track("SUBTITLE TWO", "2.2", "02:02"),
                        ],
                    ),
                    audio_track("TITLE THREE", "3"),
                ],
                1,
                [
                    ("TITLE ONE", None),
                    ("TRACK GROUP TITLE", None),
                    ("TITLE THREE", None),
                ],
                id="nested-logical-subtracks",
            ),
            _p(
                {"index_tracks": False},
                [
                    audio_track("TITLE ONE", "1"),
                    index_track(
                        "TRACK GROUP TITLE",
                        [
                            audio_track("SUBTITLE ONE", "2", "01:01"),
                            audio_track("SUBTITLE TWO", "3", "02:02"),
                        ],
                    ),
                    audio_track("TITLE FOUR", "4"),
                ],
                1,
                [
                    ("TITLE ONE", None),
                    ("SUBTITLE ONE", None),
                    ("SUBTITLE TWO", None),
                    ("TITLE FOUR", None),
                ],
                id="nested-physical-subtracks",
            ),
            _p(
                {"index_tracks": True},
                [
                    audio_track("TITLE ONE", "1"),
                    index_track(
                        "TRACK GROUP TITLE",
                        [
                            audio_track("SUBTITLE ONE", "2", "01:01"),
                            audio_track("SUBTITLE TWO", "3", "02:02"),
                        ],
                    ),
                    audio_track("TITLE FOUR", "4"),
                ],
                1,
                [
                    ("TITLE ONE", None),
                    ("TRACK GROUP TITLE: SUBTITLE ONE", None),
                    ("TRACK GROUP TITLE: SUBTITLE TWO", None),
                    ("TITLE FOUR", None),
                ],
                id="nested-physical-subtracks-with-index-tracks",
            ),
            _p(
                {"index_tracks": False},
                [
                    heading_track("MEDIUM TITLE CD1"),
                    audio_track("TITLE ONE", "1-1"),
                    audio_track("TITLE TWO", "1-2"),
                    heading_track("MEDIUM TITLE CD2"),
                    audio_track("TITLE THREE", "2-1"),
                ],
                2,
                [
                    ("TITLE ONE", "MEDIUM TITLE CD1"),
                    ("TITLE TWO", "MEDIUM TITLE CD1"),
                    ("TITLE THREE", "MEDIUM TITLE CD2"),
                ],
                id="medium-titles",
            ),
        ],
    )  # fmt: skip
    def test_parse_tracklist_index_tracks(
        self, plugin, tracks, expected_mediums, expected_tracks
    ):
        release = self._make_release(tracks)
        album = plugin.get_album_info(release)

        assert album.mediums == expected_mediums
        assert [(t.title, t.disctitle) for t in album.tracks] == expected_tracks

    def test_parse_tracklist_inherited_artists(self, plugin):
        """Verify grouped tracks combine explicit and inherited artist credits.

        This covers releases where a track group provides the default artist for
        sub-tracks that do not declare one themselves.

        Note: this is based on the following release:
        https://www.discogs.com/release/3647530
        """
        track_artist = "TRACK ARTIST"
        group_artist = "GROUP ARTIST"
        tracks = [
            index_track(
                "TRACK GROUP TITLE",
                [
                    audio_track(
                        "TRACK ONE", "2", artists=[_artist(track_artist)]
                    ),
                    audio_track("SUBTITLE TWO", "3"),
                ],
                artists=[_artist(group_artist)],
            )
        ]
        release = self._make_release(tracks)
        album = plugin.get_album_info(release)

        assert album.mediums == 1
        assert {t.artist for t in album.tracks} == {track_artist, group_artist}

    def test_coalesce_logical_index_as_audio_track(self, plugin):
        index = index_track(
            "TRACK GROUP TITLE",
            [
                audio_track("SUBTITLE ONE", "2.1"),
                audio_track("SUBTITLE TWO", "2.2"),
            ],
            duration="03:03",
        )

        assert plugin._coalesce_tracks([index]) == [
            audio_track("TRACK GROUP TITLE", "2", "03:03")
        ]


class TestDGSearchQuery(TestHelper):
    def test_default_search_filters_without_extra_tags(self):
        """Discogs search uses only the type filter when no extra_tags are set."""
        plugin = DiscogsPlugin()
        items = [Item()]

        query, filters = plugin.get_search_query_with_filters(
            "album", items, "Artist", "Album", False
        )

        assert "Album" in query
        assert filters == {"type": "release"}

    def test_extra_tags_populate_discogs_filters(self):
        """Configured extra_tags should populate Discogs search filters."""
        plugin = DiscogsPlugin()
        plugin.config["extra_tags"] = ["label", "catalognum"]

        items = [
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="def"),
        ]

        _query, filters = plugin.get_search_query_with_filters(
            "album", items, "Artist", "Album", False
        )

        assert filters["type"] == "release"
        assert filters["label"] == "abc"
        # Catalog number should have whitespace removed.
        assert filters["catno"] == "ABC123"
        config["discogs"]["extra_tags"] = []


class TestAnv:
    @pytest.fixture
    def album_info(self, monkeypatch, anv_config):
        monkeypatch.setattr(
            "beetsplug.discogs.DiscogsPlugin.setup", lambda _: None
        )
        data = {
            "id": 123,
            "uri": "https://www.discogs.com/release/123456-something",
            "tracklist": [
                {
                    "title": "track",
                    "position": "A",
                    "type_": "track",
                    "duration": "5:44",
                    "artists": [
                        _artist("ARTIST", id=11146, anv="ART", join="Feat."),
                        _artist("PERFORMER", id=787),
                    ],
                    "extraartists": [
                        _artist("PERFORMER", id=787, role="Featuring")
                    ],
                }
            ],
            "artists": [
                _artist("DRUMMER", id=445, anv="DRUM", join=", "),
                _artist("ARTIST (4)", id=321, anv="ARTY", join="&"),
                _artist("SOLOIST", id=446, anv="SOLO"),
            ],
            "title": "title",
        }
        release = get_release(data)
        plugin = DiscogsPlugin()
        plugin.config["anv"].set(
            {"artist": False, "album_artist": False, "artist_credit": False}
            | anv_config
        )
        return plugin.get_album_info(release)

    @staticmethod
    def _assert_fields(obj, expected):
        for field, value in expected.items():
            assert getattr(obj, field) == value

    @pytest.mark.parametrize(
        "anv_config,expected_track_fields",
        [
            (
                {"artist": False},
                {
                    "artist": "ARTIST Feat. PERFORMER",
                    "artists": ["ARTIST", "PERFORMER"],
                },
            ),
            (
                {"artist": True},
                {
                    "artist": "ART Feat. PERFORMER",
                    "artists": ["ART", "PERFORMER"],
                },
            ),
        ],
    )
    def test_track_artist_fields(self, album_info, expected_track_fields):
        self._assert_fields(album_info.tracks[0], expected_track_fields)

    @pytest.mark.parametrize(
        "anv_config,expected_album_fields,expected_track_fields",
        [
            (
                {"artist_credit": False},
                {
                    "artist_credit": "DRUMMER, ARTIST & SOLOIST",
                    "artists_credit": ["DRUMMER", "ARTIST", "SOLOIST"],
                },
                {
                    "artist_credit": "ARTIST Feat. PERFORMER",
                    "artists_credit": ["ARTIST", "PERFORMER"],
                },
            ),
            (
                {"artist_credit": True},
                {
                    "artist_credit": "DRUM, ARTY & SOLO",
                    "artists_credit": ["DRUM", "ARTY", "SOLO"],
                },
                {
                    "artist_credit": "ART Feat. PERFORMER",
                    "artists_credit": ["ART", "PERFORMER"],
                },
            ),
        ],
    )
    def test_artist_credit_fields(
        self, album_info, expected_album_fields, expected_track_fields
    ):
        self._assert_fields(album_info, expected_album_fields)
        self._assert_fields(album_info.tracks[0], expected_track_fields)

    @pytest.mark.parametrize(
        "anv_config,expected_album_fields",
        [
            (
                {"album_artist": False},
                {
                    "artist": "DRUMMER, ARTIST & SOLOIST",
                    "artists": ["DRUMMER", "ARTIST", "SOLOIST"],
                },
            ),
            (
                {"album_artist": True},
                {
                    "artist": "DRUM, ARTY & SOLO",
                    "artists": ["DRUM", "ARTY", "SOLO"],
                },
            ),
        ],
    )
    def test_album_artist_fields(self, album_info, expected_album_fields):
        self._assert_fields(album_info, expected_album_fields)


def test_anv_album_artist():
    """Test using artist name variations when the album artist
    is the same as the track artist, but only the track artist
    should use the artist name variation."""
    data = {
        "id": 123,
        "uri": "https://www.discogs.com/release/123456-something",
        "tracklist": [
            {
                "title": "track",
                "position": "A",
                "type_": "track",
                "duration": "5:44",
            }
        ],
        "artists": [_artist("ARTIST (4)", id=321, anv="VARIATION")],
        "title": "title",
    }
    release = get_release(data)
    config["discogs"]["anv"]["album_artist"] = False
    config["discogs"]["anv"]["artist"] = True
    config["discogs"]["anv"]["artist_credit"] = False
    r = DiscogsPlugin().get_album_info(release)
    assert r.artist == "ARTIST"
    assert r.artists == ["ARTIST"]
    assert r.artist_credit == "ARTIST"
    assert r.artist_id == "321"
    assert r.artists_credit == ["ARTIST"]
    assert r.tracks[0].artist == "VARIATION"
    assert r.tracks[0].artists == ["VARIATION"]
    assert r.tracks[0].artist_credit == "ARTIST"
    assert r.tracks[0].artists_credit == ["ARTIST"]


def test_parse_featured_artists():
    """Tests the plugins ability to parse a featured artist.
    Ignores artists that are not listed as featured."""
    track = {
        "type_": "track",
        "title": "track",
        "position": "1",
        "duration": "5:00",
        "artists": [
            _artist("NEW ARTIST", id=11146, join="&"),
            _artist("VOCALIST", id=344, join="feat."),
        ],
        "extraartists": [
            _artist("SOLOIST", id=3, role="Featuring"),
            _artist("PERFORMER (1)", id=5, role="Other Role, Featuring"),
            _artist("RANDOM", id=8, role="Written-By"),
            _artist("MUSICIAN", id=10, role="Featuring [Uncredited]"),
        ],
    }

    plugin = DiscogsPlugin()
    artistinfo = ArtistState.from_config(plugin.config, [_artist("ARTIST")])
    t, _, _ = plugin.get_track_info(track, 1, 1, artistinfo)
    assert (
        t.artist == "NEW ARTIST & VOCALIST feat. SOLOIST, PERFORMER, MUSICIAN"
    )
    assert t.artists == [
        "NEW ARTIST",
        "VOCALIST",
        "SOLOIST",
        "PERFORMER",
        "MUSICIAN",
    ]
    assert t.artists_ids == ["11146", "344", "3", "5", "10"]
    assert t.composers == ["RANDOM"]


def test_parse_extraartist_roles():
    plugin = DiscogsPlugin()
    artistinfo = ArtistState.from_config(plugin.config, [_artist("ARTIST")])
    track = {
        "type_": "track",
        "title": "track",
        "position": "1",
        "duration": "5:00",
        "artists": [_artist("TRACK ARTIST", id=11)],
        "extraartists": [
            _artist("LYRICIST", id=2, role="Lyrics By"),
            _artist("ARRANGER", id=3, role="Arranged By"),
            _artist("REMIXER", id=5, role="Remixed By"),
            _artist("COMPOSER", id=6, role="Written-By"),
        ],
    }

    t, _, _ = plugin.get_track_info(track, 1, 1, artistinfo)

    assert t.artist == "TRACK ARTIST"
    assert t.artists == ["TRACK ARTIST"]
    assert t.lyricists == ["LYRICIST"]
    assert t.arrangers == ["ARRANGER"]
    assert t.remixers == ["REMIXER"]
    assert t.composers == ["COMPOSER"]


@pytest.mark.parametrize(
    "formats, expected_media, expected_albumtype",
    [
        (None, None, None),
        (
            [
                {
                    "descriptions": ['7"', "Single", "45 RPM"],
                    "name": "Vinyl",
                    "qty": 1,
                }
            ],
            "Vinyl",
            '7", Single, 45 RPM',
        ),
    ],
)
def test_get_media_and_albumtype(formats, expected_media, expected_albumtype):
    result = DiscogsPlugin.get_media_and_albumtype(formats)

    assert result == (expected_media, expected_albumtype)


def test_va_buildartistinfo():
    config["va_name"] = "VARIOUS ARTISTS"
    expected_info = {
        "artist": "VARIOUS ARTISTS",
        "artist_id": "1",
        "artists": ["VARIOUS ARTISTS"],
        "artists_ids": ["1"],
        "artist_credit": "VARIOUS ARTISTS",
        "artists_credit": ["VARIOUS ARTISTS"],
        "arrangers": None,
        "composers": None,
        "remixers": None,
        "lyricists": None,
    }

    assert (
        ArtistState.from_config(
            DiscogsPlugin().config, [_artist("Various")]
        ).info
        == expected_info
    )


@pytest.mark.parametrize(
    "position, medium, index, subindex",
    [
        ("1", None, "1", None),
        ("A12", "A", "12", None),
        ("12-34", "12-", "34", None),
        ("CD1-1", "CD1-", "1", None),
        ("1.12", None, "1", "12"),
        ("12.a", None, "12", "A"),
        ("12.34", None, "12", "34"),
        ("1ab", None, "1", "AB"),
        # Non-standard
        ("IV", "IV", None, None),
    ],
)
def test_get_track_index(position, medium, index, subindex):
    assert DiscogsPlugin.get_track_index(position) == (medium, index, subindex)



"""Test the beets.random utilities associated with the random plugin."""

import math
import random

import pytest

from beetsplug.random import _equal_chance_permutation, random_objs


@pytest.fixture(scope="class")
def helper(class_helper):
    """Use class scope because each random test class owns its item list.

    The helper creates lightweight item objects for class setup; tests do not
    need a fresh beets temp dir per method or shared state between classes.
    """
    return class_helper


@pytest.fixture(scope="module", autouse=True)
def seed_random():
    random.seed(12345)


class TestEqualChancePermutation:
    """Test the _equal_chance_permutation function."""

    @pytest.fixture(autouse=True)
    def setup(self, helper):
        """Set up the test environment with items."""
        self.lib = helper.lib
        self.artist1 = "Artist 1"
        self.artist2 = "Artist 2"
        self.item1 = helper.create_item(artist=self.artist1)
        self.item2 = helper.create_item(artist=self.artist2)
        self.items = [self.item1, self.item2]
        for _ in range(8):
            self.items.append(helper.create_item(artist=self.artist2))

    def _stats(self, data):
        mean = sum(data) / len(data)
        stdev = math.sqrt(sum((p - mean) ** 2 for p in data) / (len(data) - 1))
        quot, rem = divmod(len(data), 2)
        if rem:
            median = sorted(data)[quot]
        else:
            median = sum(sorted(data)[quot - 1 : quot + 1]) / 2
        return mean, stdev, median

    def test_equal_permutation(self):
        """We have a list of items where only one item is from artist1 and the
        rest are from artist2. If we permute weighted by the artist field then
        the solo track will almost always end up near the start. If we use a
        different field then it'll be in the middle on average.
        """

        def experiment(field, histogram=False):
            """Permutes the list of items 500 times and calculates the position
            of self.item1 each time. Returns stats about that position.
            """
            positions = []
            for _ in range(500):
                shuffled = list(
                    _equal_chance_permutation(self.items, field=field)
                )
                positions.append(shuffled.index(self.item1))
            # Print a histogram (useful for debugging).
            if histogram:
                for i in range(len(self.items)):
                    print(f"{i:2d} {'*' * positions.count(i)}")
            return self._stats(positions)

        _, stdev1, median1 = experiment("artist")
        _, stdev2, median2 = experiment("track")
        assert 0 == pytest.approx(median1, abs=1)
        assert len(self.items) // 2 == pytest.approx(median2, abs=1)
        assert stdev2 > stdev1

    @pytest.mark.parametrize(
        "input_items, field, expected",
        [
            ([], "artist", []),
            ([{"artist": "Artist 1"}], "artist", [{"artist": "Artist 1"}]),
            # Missing field should not raise an error, but return empty
            ([{"artist": "Artist 1"}], "nonexistent", []),
            # Multiple items with the same field value
            (
                [{"artist": "Artist 1"}, {"artist": "Artist 1"}],
                "artist",
                [{"artist": "Artist 1"}, {"artist": "Artist 1"}],
            ),
        ],
    )
    def test_equal_permutation_items(
        self, input_items, field, expected, helper
    ):
        """Test _equal_chance_permutation with empty input."""
        result = list(
            _equal_chance_permutation(
                [helper.create_item(**i) for i in input_items], field
            )
        )

        for item in expected:
            for key, value in item.items():
                assert any(getattr(r, key) == value for r in result)
        assert len(result) == len(expected)


class TestRandomObjs:
    """Test the random_objs function."""

    @pytest.fixture(autouse=True)
    def setup(self, helper):
        """Set up the test environment with items."""
        self.lib = helper.lib
        self.artist1 = "Artist 1"
        self.artist2 = "Artist 2"
        self.items = [
            helper.create_item(artist=self.artist1, length=180),  # 3 minutes
            helper.create_item(artist=self.artist2, length=240),  # 4 minutes
            helper.create_item(artist=self.artist2, length=300),  # 5 minutes
        ]

    def test_random_selection_by_count(self):
        """Test selecting a specific number of items."""
        selected = list(random_objs(self.items, "artist", number=2))
        assert len(selected) == 2
        assert all(item in self.items for item in selected)

    def test_random_selection_by_time(self):
        """Test selecting items constrained by total time (minutes)."""
        selected = list(
            random_objs(self.items, "artist", time_minutes=6)
        )  # 6 minutes
        total_time = (
            sum(item.length for item in selected) / 60
        )  # Convert to minutes
        assert total_time <= 6

    def test_equal_chance_permutation(self, helper):
        """Test equal chance permutation ensures balanced artist selection."""
        # Add more items to make the test meaningful
        for _ in range(5):
            self.items.append(
                helper.create_item(artist=self.artist1, length=180)
            )

        selected = list(
            random_objs(self.items, "artist", number=10, equal_chance=True)
        )
        artist_counts = {}
        for item in selected:
            artist_counts[item.artist] = artist_counts.get(item.artist, 0) + 1

        # Ensure both artists are represented (not strictly equal due to randomness)
        assert len(artist_counts) >= 2

    def test_empty_input_list(self):
        """Test behavior with an empty input list."""
        selected = list(random_objs([], "artist", number=1))
        assert len(selected) == 0

    def test_no_constraints_returns_all(self):
        """Test that no constraints return all items in random order."""
        selected = list(random_objs(self.items, "artist", number=3))
        assert len(selected) == len(self.items)
        assert set(selected) == set(self.items)



import pytest

from beets.util.config import UnknownPairError, sanitize_choices, sanitize_pairs


@pytest.mark.parametrize(
    "input_choices, valid_choices, expected",
    [
        (["A", "Z"], ("A", "B"), ["A"]),
        (["A", "A"], ("A"), ["A"]),
        (["D", "*", "A"], ("A", "B", "C", "D"), ["D", "B", "C", "A"]),
    ],
)
def test_sanitize_choices(input_choices, valid_choices, expected):
    assert sanitize_choices(input_choices, valid_choices) == expected


def test_sanitize_pairs():
    assert sanitize_pairs(
        [
            ("foo", "baz bar"),
            ("foo", "baz bar"),
            ("key", "*"),
            ("*", "*"),
            ("discard", "bye"),
        ],
        [("foo", "bar"), ("foo", "baz"), ("foo", "foobar"), ("key", "value")],
    ) == [("foo", "baz"), ("foo", "bar"), ("key", "value"), ("foo", "foobar")]


def test_sanitize_pairs_unknown_key():
    with pytest.raises(UnknownPairError):
        sanitize_pairs(
            [("bar", "foo")], [("key", "value")], raise_on_unknown=True
        )


def test_sanitize_pairs_unknown_key_wildcard_value():
    with pytest.raises(UnknownPairError):
        sanitize_pairs(
            [("foo", "*")], [("key", "value")], raise_on_unknown=True
        )



"""Provides utilities to read, write and manipulate m3u playlist files."""

import traceback

from beets.util import FilesystemError, mkdirall, normpath, syspath


class EmptyPlaylistError(Exception):
    """Raised when a playlist file without media files is saved or loaded."""


class M3UFile:
    """Reads and writes m3u or m3u8 playlist files."""

    def __init__(self, path):
        """``path`` is the absolute path to the playlist file.

        The playlist file type, m3u or m3u8 is determined by 1) the ending
        being m3u8 and 2) the file paths contained in the list being utf-8
        encoded. Since the list is passed from the outside, this is currently
        out of control of this class.
        """
        self.path = path
        self.extm3u = False
        self.media_list = []

    def load(self):
        """Reads the m3u file from disk and sets the object's attributes."""
        pl_normpath = normpath(self.path)
        try:
            with open(syspath(pl_normpath), "rb") as pl_file:
                raw_contents = pl_file.readlines()
        except OSError as exc:
            raise FilesystemError(
                exc, "read", (pl_normpath,), traceback.format_exc()
            )

        self.extm3u = True if raw_contents[0].rstrip() == b"#EXTM3U" else False
        for line in raw_contents[1:]:
            if line.startswith(b"#"):
                # Support for specific EXTM3U comments could be added here.
                continue
            self.media_list.append(normpath(line.rstrip()))
        if not self.media_list:
            raise EmptyPlaylistError

    def set_contents(self, media_list, extm3u=True):
        """Sets self.media_list to a list of media file paths.

        Also sets additional flags, changing the final m3u-file's format.

        ``media_list`` is a list of paths to media files that should be added
        to the playlist (relative or absolute paths, that's the responsibility
        of the caller). By default the ``extm3u`` flag is set, to ensure a
        save-operation writes an m3u-extended playlist (comment "#EXTM3U" at
        the top of the file).
        """
        self.media_list = media_list
        self.extm3u = extm3u

    def write(self):
        """Writes the m3u file to disk.

        Handles the creation of potential parent directories.
        """
        header = [b"#EXTM3U"] if self.extm3u else []
        if not self.media_list:
            raise EmptyPlaylistError
        contents = header + self.media_list
        pl_normpath = normpath(self.path)
        mkdirall(pl_normpath)

        try:
            with open(syspath(pl_normpath), "wb") as pl_file:
                for line in contents:
                    pl_file.write(line + b"\n")
                pl_file.write(b"\n")  # Final linefeed to prevent noeol file.
        except OSError as exc:
            raise FilesystemError(
                exc, "create", (pl_normpath,), traceback.format_exc()
            )



"""Provides the basic, interface-agnostic workflow for importing and
autotagging music files.
"""

from .actions import Action, DuplicateAction
from .session import ImportAbortError, ImportSession
from .tasks import (
    ArchiveImportTask,
    ImportTask,
    SentinelImportTask,
    SingletonImportTask,
)

# Note: Stages are not exposed to the public API

__all__ = [
    "Action",
    "ArchiveImportTask",
    "DuplicateAction",
    "ImportAbortError",
    "ImportSession",
    "ImportTask",
    "SentinelImportTask",
    "SingletonImportTask",
]



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Collection, Sequence


def sanitize_choices(
    choices: Sequence[str], choices_all: Collection[str]
) -> list[str]:
    """Clean up a stringlist configuration attribute: keep only choices
    elements present in choices_all, remove duplicate elements, expand '*'
    wildcard while keeping original stringlist order.
    """
    seen: set[str] = set()
    others = [x for x in choices_all if x not in choices]
    res: list[str] = []
    for s in choices:
        if s not in seen:
            if s in list(choices_all):
                res.append(s)
            elif s == "*":
                res.extend(others)
        seen.add(s)
    return res


def sanitize_pairs(
    pairs: Sequence[tuple[str, str]],
    pairs_all: Sequence[tuple[str, str]],
    raise_on_unknown: bool = False,
) -> list[tuple[str, str]]:
    """Clean up a single-element mapping configuration attribute as returned
    by Confuse's `Pairs` template: keep only two-element tuples present in
    pairs_all, remove duplicate elements, expand ('str', '*') and ('*', '*')
    wildcards while keeping the original order. Note that ('*', '*') and
    ('*', 'whatever') have the same effect.
    Set raise_on_unknown to raise an error when a provided pair is not recognised

    For example,

    >>> sanitize_pairs(
    ...     [('foo', 'baz bar'), ('key', '*'), ('*', '*')],
    ...     [('foo', 'bar'), ('foo', 'baz'), ('foo', 'foobar'),
    ...      ('key', 'value')]
    ...     )
    [('foo', 'baz'), ('foo', 'bar'), ('key', 'value'), ('foo', 'foobar')]
    """
    pairs_all = list(pairs_all)
    seen: set[tuple[str, str]] = set()
    others = [x for x in pairs_all if x not in pairs]
    res: list[tuple[str, str]] = []
    for k, values in pairs:
        for v in values.split():
            x = (k, v)
            if x in pairs_all:
                if x not in seen:
                    seen.add(x)
                    res.append(x)
            elif k == "*":
                new = [o for o in others if o not in seen]
                seen.update(new)
                res.extend(new)
            elif v == "*":
                new = [o for o in others if o not in seen and o[0] == k]

                if len(new) == 0 and raise_on_unknown:
                    raise UnknownPairError(k, v)

                seen.update(new)
                res.extend(new)
            elif raise_on_unknown:
                raise UnknownPairError(k, v)
    return res


class UnknownPairError(Exception):
    def __init__(self, k, v):
        super().__init__(f"setting {k}={v} is not recognized")



"""A simple utility for constructing filesystem-like trees from beets
libraries.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple

from beets import util

if TYPE_CHECKING:
    from beets.library import Library


class Node(NamedTuple):
    files: dict[str, int]
    # Maps filenames to Item ids.

    dirs: dict[str, Node]
    # Maps directory names to child nodes.


def _insert(node: Node, path: list[str], itemid: int):
    """Insert an item into a virtual filesystem node."""
    if len(path) == 1:
        # Last component. Insert file.
        node.files[path[0]] = itemid
    else:
        # In a directory.
        dirname = path[0]
        rest = path[1:]
        if dirname not in node.dirs:
            node.dirs[dirname] = Node({}, {})
        _insert(node.dirs[dirname], rest, itemid)


def libtree(lib: Library) -> Node:
    """Generates a filesystem-like directory tree for the files
    contained in `lib`. Filesystem nodes are (files, dirs) named
    tuples in which both components are dictionaries. The first
    maps filenames to Item ids. The second maps directory names to
    child node tuples.
    """
    root = Node({}, {})
    for item in lib.items():
        if item.id is None:
            continue
        dest = item.destination(relative_to_libdir=True)
        parts = util.components(util.as_string(dest))
        _insert(root, parts, item.id)
    return root
